"""
    domonic.d3.selection
    ====================================

    https://github.com/d3/d3-selection/tree/main/src/selection

"""

# from domonic.html import *

from domonic.dom import document  # bring in the global
from domonic.javascript import *

xhtml = "http://www.w3.org/1999/xhtml"

namespaces = {"svg": "http://www.w3.org/2000/svg", "xhtml": xhtml, "xlink": "http://www.w3.org/1999/xlink", "xml": "http://www.w3.org/XML/1998/namespace", "xmlns": "http://www.w3.org/2000/xmlns/"}

# export {default as namespace} from "./namespace.js";
# export {default as namespaces} from "./namespaces.js";
def namespace(name):
    name = str(name)
    prefix = name
    i = String(prefix).indexOf(":")
    if i > 0:
        prefix = String(name).slice(0, i)
    if i >= 0 and prefix != "xmlns":
        name = String(name).slice(i + 1)
    return {"space": namespaces[prefix], "local": name} if Object(namespaces).hasOwnProperty(prefix) else name  # eslint-disable-line no-prototype-builtins


def creatorInherit(name):
    def anon(this):
        # document = this.ownerDocument
        from domonic.dom import document

        # uri = this.namespaceURI
        uri = document.namespaceURI
        return document.createElement(name) if uri == xhtml and document.documentElement.namespaceURI == xhtml else document.createElementNS(uri, name)
    return anon


def creatorFixed(fullname):
    # print('this one')
    # return lambda this: print(this[0] , "TESTESTESTE")
    # return lambda this: this.ownerDocument.createElementNS(fullname['space'], fullname['local'])
    from domonic.dom import document  # bring in the global document
    return lambda *args: document.ownerDocument.createElementNS(fullname['space'], fullname['local'])


def creator(name):
    fullname = namespace(name)
    # print(fullname)
    func = creatorFixed if isinstance(fullname, dict) else creatorInherit
    return func(fullname)


def none():
    return {}


def selector(selector):
    return None if selector == None else lambda: document.querySelector(selector)

# // Given something array like (or null), returns something that is strictly an
# // array. This is used to ensure that array-like objects passed to d3.selectAll
# // or selection.selectAll are converted into proper arrays when creating a
# // selection; we don’t ever want to create a selection backed by a live
# // HTMLCollection or NodeList. However, note that selection.selectAll will use a
# // static NodeList as a group, since it safely derived from querySelectorAll.
def array(x):
    b = x if Array.isArray(x) else Array.from_(x)
    return [] if x == None else b


# export {default as window} from "./window.js";
def window(node):
    return (node.ownerDocument and node.ownerDocument.defaultView) or (node.document and node) or node.defaultView


defaultView = window

# import selection_select from "./select.js";

# import selection_style from "./style.js";
def styleValue(node, name):
    return node.style.getPropertyValue(name) or defaultView(node).getComputedStyle(node, None).getPropertyValue(name)


def sparse(self, update):
    return Array(len(update))


class EnterNode():

    def __init__(self, parent, datum):
        self.ownerDocument = parent.ownerDocument
        self.namespaceURI = parent.namespaceURI
        self._next = None
        self._parent = parent
        self.__data__ = datum

    def appendChild(self, child):
        return self._parent.insertBefore(child, self._next)

    def insertBefore(self, child, next):
        return self._parent.insertBefore(child, next)

    def querySelector(self, selector):
        return self._parent.querySelector(selector)

    def querySelectorAll(self, selector):
        return self._parent.querySelectorAll(selector)


class ClassList():

    def __init__(self, node):
        # print('class list is in town')
        self._node = node
        self._names = classArray(node.getAttribute("class") or "")

    def add(self, name):
        i = String(self._names).indexOf(name)
        if i < 0:
            self._names.append(name)
            self._node.setAttribute("class", " ".join(self._names))

    def remove(self, name):
        i = self._names.indexOf(name)
        if i >= 0:
            self._names.splice(i, 1)
            self._node.setAttribute("class", " ".join(self._names))

    def contains(self, name):
        return self._names.indexOf(name) >= 0


def classArray(string):
    return String(string).trim().split(r'/^|\s+/')


def classList(node):
    # return node.classList or ClassList(node)
    return ClassList(node)


def classedAdd(node, names):
    mylist = classList(node)
    i = -1
    n = len(names)
    while i < n:
        mylist.add(names[i])
        i += 1


def classedRemove(node, names):
    list = classList(node)
    i = -1
    n = len(names)
    while i < n:
        list.remove(names[i])
        i += 1


# import selection_append from "./append.js";


root = [None]


class Selection():

    def __init__(self, groups, parents, this=None):
        self._groups = groups
        self._parents = parents
        self.this = None
        if this is None:
            # self.this = root[0]
            # self.this = self._groups#[0]
            # self.this = self._groups  #.__iter__()
            pass
        else:
            self.this = this  # context switcher

    # unpack groups into a list of nodes
    def __iter__(self):
        return self._groups.__iter__()

    def select(self, select):
        if not callable(select):
            select = selector(select)

        groups = self._groups
        m = len(groups)
        subgroups = Array(m)
        j = 0
        for group in groups:
            n = len(group)
            subgroup = subgroups[j] = Array(n)
            for i in range(n):
                node = group[i]
                if node is None:
                    print('NODE WAS NONE.err?')
                    continue
                try:
                    # print(node.__data__)
                    node.__data__ = None
                    # print('bipm', select)
                    subnode = Function(select).call(node, node.__data__, i, group)
                except Exception as e:
                    # print(e)
                    print('failed. no __data__ on node')
                    subnode = None
                # if subnode is not None:
                #     if "__data__" in subnode:
                #         subnode.__data__ = subnode.__data__
                #     subgroup[i] = subnode
                    # subnode.__data__ = node.__data__
                    # subgroup[i] = subnode
                # print('super::', node, subnode)
                if "__data__" in node:
                    subnode.__data__ = node.__data__
                subgroup[i] = subnode
            j += 1

        # print("this was set to", self.this)
        return Selection(subgroups, self._parents, self.this)

    # import selection_selectAll from "./selectAll.js";
    # import {Selection} from "./index.js";
    # import array from "../array.js";
    # import selectorAll from "../selectorAll.js";

    def arrayAll(self, select, *args):
        return lambda: array(Function(select).apply(self.this, args))

    def selectAll(self, select):
        if callable(select):
            select = self.arrayAll(select)
        else:
            select = selectorAll(select)

        groups = self._groups
        m = len(groups)
        subgroups = []
        parents = []
        j = 0
        for group in groups:
            n = len(group)
            for i in range(n):
                node = group[i]
                if node is None:
                    print('selectaAll : NODE WAS NONE.err?')
                    continue

                try:
                    # print(node.__data__)
                    node.__data__ = None  # TODO - only do this if not there
                    subgroups.append(Function(select).call(node, node.__data__, i, group))
                    parents.append(node)
                except Exception as e:
                    # print(e)
                    print('failed. no __data__ on node')

                # subgroups.append(Function(select).call(node, node.__data__, i, group))
                # parents.append(node)
            j += 1
        return Selection(subgroups, parents, self.this)

    # import selection_selectChild from "./selectChild.js";
    # import {childMatcher} from "../matcher.js";
    find = Array.prototype.find

    def childFind(self, match):
        return lambda: find.call(this.children, match)

    def childFirst(self):
        return this.firstElementChild

    def selectChild(self, match):
        # return this.select(match == null ? childFirst : childFind(typeof match === "function" ? match : childMatcher(match)))
        if callable(match):
            match = childFind(match)
        else:
            match = childMatcher(match)
        return this.select(match)

    # import selection_selectChildren from "./selectChildren.js";
    # def selectChildren: selection_selectChildren,
    # import {childMatcher} from "../matcher.js";

    # filter = Array.prototype.filter

    def children(self):
        return Array.from_(self.this.children)

    def childrenFilter(self, match):
        return lambda: Array.filter.call(self.this.children, match)

    def selectChildren(self, match):
        # TODO - rewrite this as python
        # return this.selectAll(match == None ? children : childrenFilter(typeof match === "function" ? match : childMatcher(match)));
        # return self.selectAll(match == None ? self.children : self.childrenFilter(typeof match == "function" ? match : childMatcher(match)))
        if match is None:
            return self.selectAll(self.this.children)
        else:
            return self.selectAll(self.childrenFilter(match))

    # import selection_filter from "./filter.js";
    # def filter: selection_filter,
    # import {Selection} from "./index.js";
    # import matcher from "../matcher.js"; # TODO - might not be in yet

    def filter(self, match):
        if callable(match):
            match = matcher(match)

        # TODO - rewrite this as python
        # for (var groups = this._groups, m = groups.length, subgroups = new Array(m), j = 0; j < m; ++j) {
        #     for (var group = groups[j], n = group.length, subgroup = subgroups[j] = [], node, i = 0; i < n; ++i) {
        #     if ((node = group[i]) && match.call(node, node.__data__, i, group)) {
        #         subgroup.push(node);
        #     }
        #     }
        # }
        groups = self._groups
        m = len(groups)
        subgroups = []
        j = 0
        for group in groups:
            n = len(group)
            for i in range(n):
                node = group[i]
                if node is None:
                    continue
                if match.call(node, node.__data__, i, group):
                    subgroups.append(node)
            j += 1
        return Selection(subgroups, self._parents, self.this)

    # import selection_data from "./data.js";
    # def data: selection_data, # TODO -------------------------------------- this is a big one

    # import selection_enter from "./enter.js";
    # def enter: selection_enter,
    # import sparse from "./sparse.js";
    # import {Selection} from "./index.js";

    def enter(self):
        return Selection(self._enter or self._groups.map(sparse), self._parents)

    # import selection_exit from "./exit.js";
    # def exit: selection_exit,
    # import sparse from "./sparse.js";
    # import {Selection} from "./index.js";
    def exit(self):
        return Selection(this._exit or self._groups.map(sparse), self._parents)

    # import selection_join from "./join.js";
    # def join: selection_join,
    def join(self, onenter, onupdate, onexit):
        enter = this.enter()
        update = this
        exit = self.exit()
        if callable(onenter):
            enter = onenter(enter)
            if (enter):
                enter = enter.selection()
        else:
            enter = enter.append(onenter + "")

        if onupdate != None:
            update = onupdate(update)
            if (update):
                update = update.selection()
        if onexit == None:
            exit.remove()
        else:
            onexit(exit)
        return enter.merge(update).order() if enter and update else update

    # import selection_merge from "./merge.js";
    # def merge: selection_merge,
    # import {Selection} from "./index.js";
    def merge(self, context):
        selection = context.selection() if context.selection else context

        # TODO - rewrite this as python
        # for (var groups0 = this._groups, groups1 = selection._groups, m0 = groups0.length, m1 = groups1.length, m = Math.min(m0, m1), merges = new Array(m0), j = 0; j < m; ++j) {
        #     for (var group0 = groups0[j], group1 = groups1[j], n = group0.length, merge = merges[j] = new Array(n), node, i = 0; i < n; ++i) {
        #     if (node = group0[i] || group1[i]) {
        #         merge[i] = node;
        #     }
        #     }
        # }
        groups0 = self._groups
        groups1 = context.selection._groups
        m0 = len(groups0)
        m1 = len(groups1)
        m = min(m0, m1)
        merges = []
        j = 0
        for group0 in groups0:
            group1 = groups1[j]
            n = len(group0)
            merge = []
            for i in range(n):
                node = group0[i]
                if node is None:
                    continue
                if group1[i] is None:
                    continue
                merge.append(node)
            merges.append(merge)
            j += 1

        # TODO - rewrite this as python
        # for (; j < m0; ++j) {
        #     merges[j] = groups0[j];
        # }
        while j < m0:
            merges[j] = groups0[j]
            j += 1

        return Selection(merges, self._parents, self.this)


    # def selection: selection_selection, # ---?? TODO - is this right?

    # import selection_order from "./order.js";
    # def order: selection_order,
    def order(self):
        # TODO - rewrite this as python
        # for (var groups = this._groups, j = -1, m = groups.length; ++j < m;) {
        #     for (var group = groups[j], i = group.length - 1, next = group[i], node; --i >= 0;) {
        #     if (node = group[i]) {
        #         if (next && node.compareDocumentPosition(next) ^ 4) next.parentNode.insertBefore(node, next);
        #         next = node;
        #     }
        #     }
        # }
        groups = self._groups
        j = -1
        m = len(groups)
        while j < m:
            i = len(groups[j]) - 1
            next = groups[j][i]
            while i >= 0:
                node = groups[j][i]
                if node is None:
                    continue
                if next and node.compareDocumentPosition(next) ^ 4:
                    next.parentNode.insertBefore(node, next)
                next = node
                i -= 1
            j += 1
        return self

    # import selection_sort from "./sort.js";
    # def sort: selection_sort,
    # import {Selection} from "./index.js";
    def sort(self, compare):
        if not compare:
            compare = ascending

        def compareNode(a, b):
            # TODO - rewrite this as python
            # return a and b ? compare(a.__data__, b.__data__) : !a - !b
            if a and b:
                return compare(a.__data__, b.__data__)
            return -1 if a else 1

        # TODO - rewrite this as python
        # for (var groups = this._groups, m = groups.length, sortgroups = new Array(m), j = 0; j < m; ++j) {
        #     for (var group = groups[j], n = group.length, sortgroup = sortgroups[j] = new Array(n), node, i = 0; i < n; ++i) {
        #     if (node = group[i]) {
        #         sortgroup[i] = node;
        #     }
        #     }
        #     sortgroup.sort(compareNode);
        # }
        groups = self._groups
        m = len(groups)
        sortgroups = []
        j = 0
        for group in groups:
            n = len(group)
            sortgroup = []
            for i in range(n):
                node = group[i]
                if node is None:
                    continue
                sortgroup.append(node)
            sortgroup.sort(compareNode)
            sortgroups.append(sortgroup)
            j += 1
        return Selection(sortgroups, self._parents, self.this).order()

    def ascending(self, a, b):
        # return a < b ? -1 : a > b ? 1 : a >= b ? 0 : NaN
        if a < b:
            return -1
        if a > b:
            return 1
        if a >= b:
            return 0
        return None

    # import selection_call from "./call.js"
    def call(self, *args):
        args = list(args)
        callback = args[0]
        args[0] = self
        Function(callback).apply(None, args)
        return self

    # import selection_nodes from "./nodes.js";
    def nodes(self):
        return Array.from_(self.this)

    # import selection_node from "./node.js";
    def node(self):
        groups = self._groups
        j = 0
        m = len(groups)
        for group in groups:
            i = 0
            n = len(group)
            while i < n:
                node = group[i]
                if node:
                    return node
                i += 1
        return None

    # import selection_size from "./size.js";
    # def size: selection_size,
    def size(self):
        size = 0
        for node in self.this:
            size += 1  # eslint-disable-line no-unused-vars
        return size

    # import selection_empty from "./empty.js";
    # def empty: selection_empty,
    def empty(self):
        # return not self.this.node()
        if isinstance(self.this, list):
            return not self.this

        if self.node() is None:
            return True
        return False
        # return not self.this.node()

    # import selection_each from "./each.js";
    def each(self, callback):
        groups = self._groups
        j = 0
        m = len(groups)
        for group in groups:
            i = 0
            n = len(group)
            while i < n:
                node = group[i]
                if node is None:
                    print('NODE WAS NONE.err?')
                    continue
                # try:
                node.__data__ = None
                Function(callback).call(node, node.__data__, i, group)
                # print('worked on this one')
                # except Exception as e:
                    # print(e)
                    # print('failed. no __data__ on node mate', e)

                i += 1
            j += 1
        return self

    # import selection_attr from "./attr.js";
    def attrRemove(self, name):
        # return lambda this: this.removeAttribute(name)
        self.this.removeAttribute(name)
        return self

    def attrRemoveNS(self, fullname):
        # return lambda this: this.removeAttributeNS(fullname['space'], fullname['local'])
        self.this.removeAttributeNS(fullname['space'], fullname['local'])
        return self

    def attrConstant(self, name, value):
        # print('setting:::', name, value)
        # print('setting:::', self.this)
        # return lambda: self.setAttribute(name, value)
        # self.this.setAttribute(name, value)
        # return self

        def anon(this, *args):
            nonlocal name
            nonlocal value
            # this.textContent = value
            return this.setAttribute(name, value)

        return anon

    def attrConstantNS(self, fullname, value):
        # return lambda this: this.setAttributeNS(fullname['space'], fullname['local'], value)
        self.this.setAttributeNS(fullname['space'], fullname['local'], value)
        return self

    def attrFunction(self, name, value, *args):
        def anon(this):
            nonlocal value
            nonlocal name
            nonlocal args
            v = Function(value).apply(this, args)
            if v == None:
                this.removeAttribute(name)
            else:
                this.setAttribute(name, v)

        return self

    def attrFunctionNS(self, fullname, value, *args):
        def anon(this):
            nonlocal value
            nonlocal fullname
            nonlocal args
            v = Object(value).apply(this, args)
            if v == None:
                this.removeAttributeNS(fullname['space'], fullname['local'])
            else:
                this.setAttributeNS(fullname['space'], fullname['local'], v)
        return anon

    def attr(self, name, value, *args):
        # print("NAME!!", name, value, args)
        fullname = namespace(name)

        # if len(args) > 0: #4:
        if value is None:
            node = self.node()
            print(node)
            return node.getAttributeNS(fullname['space'], fullname['local']) if isinstance(fullname, dict) else node.getAttribute(fullname)

        a = self.attrRemoveNS if getattr(fullname, 'local', None) is not None else self.attrRemove
        b = self.attrFunctionNS if getattr(fullname, 'local', None) is not None else self.attrFunction
        c = self.attrConstantNS if getattr(fullname, 'local', None) is not None else self.attrConstant

        if value is None:
            func = a
        elif callable(value):
            func = b
        else:
            func = c

        self.each(func(fullname, value))
        return self

    # def style: selection_style,
    # import defaultView from "../window.js";
    def _styleRemove(self, name, value, priority=None):
        # print('styleing remove')
        def anon(this, *args):
            # print('_styleRemove :anon/name', name)
            this.style.removeProperty(name)
        return anon

    def _styleConstant(self, name, value, priority=None):
        # print('style constant was called')
        def anon(this, *args):
            # print('THE FUNC WAS CALLED')
            nonlocal name
            nonlocal value
            nonlocal priority
            # print('_styleConstantxxx :anon/name', name, type(this), this)
            # print('aaa',this)
            # print('aaaaawtf')
            # print('bbb',this.style)
            # print('ccc')
            this.style.setProperty(name, value, priority)
        return anon

    def _styleFunction(self, name, value, priority=None):
        # print('styling fucntion')
        def anon(this, *args):
            # print('styling fucntion:anon/name', name)
            v = Function(value).apply(this, args)
            # print('dark mavis',v)
            if v == None:
                this.style.removeProperty(name)
            else:
                # print('how you doin')
                this.style.setProperty(name, v, priority)
        return anon

    def style(self, name, value=None, priority=None, *args):
        if value == None:
            return styleValue(self.node(), name)

        # print('hi!!!!!!!!')
        if value == None:  # ?? need to understand what below is doing
            func = self._styleRemove  # (name, value, priority)
        elif callable(value):
            func = self._styleFunction  # (name, value, priority)
        else:
            func = self._styleConstant  # (name, value, priority)

        p = "" if priority == None else priority
        return self.each(func(name, value, p))


    def append(self, name, *args):
        create = name if callable(name) else creator(name)
        # print(create)

        def anon(this, *args):
            # print("THIS", this, args)
            nonlocal create
            # nonlocal self
            # print("self is::", self)
            n = Function(create).apply(self, args)
            # print('n is::', n)
            # print("self this is::", self.this)
            # self.this = this
            return this.appendChild(n)

        self.select(anon)
        return self

    # import selection_property from "./property.js";
    def propertyRemove(self, name):
        def anon(this):
            del this[name]
        return anon

    def propertyConstant(self, name, value):
        def anon(this):
            this[name] = value
        return anon

    def propertyFunction(self, name, value):
        def anon(this, *args):
            v = Function(value).apply(this, args)
            if v == None:
                del this[name]
            else:
                this[name] = v
        return anon

    def property(self, name, value):

    # TODO write this commented out javascript as python instead
    # return arguments.length > 1
    #     ? this.each((value == null
    #         ? propertyRemove : typeof value === "function"
    #         ? propertyFunction
    #         : propertyConstant)(name, value))
    #     : this.node()[name]
        if value == None:
            func = propertyRemove
        elif callable(value):
            func = propertyFunction
        else:
            func = propertyConstant

        return func(name, value)

    # import selection_text from "./text.js";
    # import selection_classed from "./classed.js";

    # def classed: selection_classed,

    def classedTrue(self, names, value):
        # print("classedTrue::::")
        return lambda this, *args: classedAdd(this, names)

    def classedFalse(self, names, value):
        return lambda this, *args: classedRemove(this, names)

    def classedFunction(self, names, value):
        # TODO write this commented out javascript as python instead
        # return function() {
        # (value.apply(this, arguments) ? classedAdd : classedRemove)(this, names)
        # }
        def anon(this, *args):
            nonlocal names
            nonlocal value
            # nonlocal this
            # nonlocal args
            # print('classedFunction', names, value)
            v = Function(value).apply(this, args)
            if v == None:
                classedRemove(this, names)
            else:
                classedAdd(this, names)
        return anon

    def classed(self, name, value, *args):
        names = classArray(str(name))
        # print(names)
        # if (args.length < 2):
        if value == None:
            list = classList(this.node())
            i = -1
            n = len(names)
            while (i < n):
                if not list.contains(names[i]):
                    return False
                i += 1
            return True

        # TODO write this commented out javascript as python instead
        # return this.each((typeof value === "function"
        #     ? classedFunction : value
        #     ? classedTrue
        #     : classedFalse)(names, value));
        # }
        # print( 'hey',value, callable(value) )

        if value == None:
            func = self.classedFalse
        elif callable(value):
            func = self.classedFunction
        else:
            func = self.classedTrue

        self.each(func(names, value))


    # def text: selection_text,
    def _textRemove(self):
        self.this.textContent = ""

    def _textConstant(self, value):
        def anon(this, *args):
            this.textContent = value
        return anon

    def _textFunction(self, value):
        def anon(this, *args):
            v = Function(value).apply(this, args)
            this.textContent = "" if v == None else v
        return anon

    def text(self, value=None):
        if value == None:
            return self.node()._textContent
        # return arguments.length
        #     ? this.each(value == null
        #         ? textRemove : (typeof value === "function"
        #         ? textFunction
        #         : textConstant)(value))
        #     : this.node().textContent
        if value == None:
            func = self._textRemove
        elif callable(value):
            func = self._textFunction
        else:
            func = self._textConstant

        self.each(func(value))

    # import selection_html from "./html.js";
    def htmlRemove(self):
        self.this.innerHTML = ""

    def htmlConstant(self, value):
        def anon(this):
            this.innerHTML = value
        return anon

    def htmlFunction(self, value):
        def anon(this, *args):
            v = Function(value).apply(this, args)
            this.innerHTML = "" if v == None else v
        return anon

    def html(self, value):
        #TODO write this commented out javascript as python instead
        # return arguments.length
        #     ? this.each(value == null
        #         ? htmlRemove : (typeof value === "function"
        #         ? htmlFunction
        #         : htmlConstant)(value))
        #     : this.node().innerHTML;
        if value == None:
            func = htmlRemove
        elif callable(value):
            func = htmlFunction
        else:
            func = htmlConstant
        return func(value)


    # import selection_raise from "./raise.js";
    # def _raise(self):
    #     if (this.nextSibling):
    #         this.parentNode.appendChild(this)

    # def raise(self):
    #     return this.each(raise)

    # import selection_lower from "./lower.js";
    # def lower: selection_lower,
    def _lower(self):
        if self.this.previousSibling:
            self.this.parentNode.insertBefore(self.this, self.this.parentNode.firstChild)

    def lower(self):
        return self.each(lower)

    # import selection_insert from "./insert.js";
    # import creator from "../creator.js"; # already in?
    # import selector from "../selector.js"; # already in?

    def constantNull(self):
        return None

    def insert(self, name, before, *args):
        # TODO write this commented out javascript as python instead
        # var create = typeof name === "function" ? name : creator(name),
        #     select = before == null ? constantNull : typeof before === "function" ? before : selector(before);
        # return this.select(function() {
        #     return this.insertBefore(create.apply(this, arguments), select.apply(this, arguments) || null);
        # })
        create = name if callable(name) else creator(name)
        select = before == None or before == "null" or before == "undefined" or before == "null"
        return self.select(lambda: self.this.insertBefore(Function(create).apply(self, args), select))

    # import selection_remove from "./remove.js";
    # def remove: selection_remove,
    def _remove(self):
        parent = self.this.parentNode
        if parent:
            parent.removeChild(self.this)

    def remove(self):
        return self.each(remove)

    # import selection_clone from "./clone.js";
    def selection_cloneShallow(self):
        clone = this.cloneNode(False)
        parent = self.this.parentNode
        return parent.insertBefore(clone, this.nextSibling) if parent else clone

    def selection_cloneDeep(self):
        clone = this.cloneNode(False)
        parent = self.this.parentNode
        return parent.insertBefore(clone, self.this.nextSibling) if parent else clone

    def clone(self, deep):
        return this.select(selection_cloneDeep if deep else selection_cloneShallow)

    # import selection_datum from "./datum.js";
    def datum(self, value=None, *args):
        # return arguments.length
        #     ? this.property("__data__", value)
        #     : this.node().__data__;
        # }
        return self.this.property("__data__", value) if value is not None else self.node().__data__

    # import selection_on from "./on.js";
    def contextListener(self, listener):
        return lambda event: listener.call(self.this, event, self.this.__data__)

    def parseTypenames(self, typenames):
        # TODO - write this as python
        # return typenames.trim().split(/^|\s+/).map(function(t) {
        #     var name = "", i = t.indexOf(".");
        #     if (i >= 0) name = t.slice(i + 1), t = t.slice(0, i)
        #     return {type: t, name: name}
        # });
        return [{'type': t[0], 'name': t[1]} for t in re.findall(r'\.([^\.]+)', typenames)]

    def onRemove(self, typename):
        # TODO - write this as python
        # return function() {
        #     var on = this.__on;
        #     if (!on) return;
        #     for (var j = 0, i = -1, m = on.length, o; j < m; ++j) {
        #     if (o = on[j], (!typename.type || o.type === typename.type) && o.name === typename.name) {
        #         this.removeEventListener(o.type, o.listener, o.options);
        #     } else {
        #         on[++i] = o;
        #     }
        #     }
        #     if (++i) on.length = i;
        #     else delete this.__on;
        # }
        def anon(this):
            on = this.__on
            if not on:
                return
            for j in range(0, len(on)):
                o = on[j]
                if not typename:
                    on[j] = o
                    return
                if o.type and o.type == typename.type:
                    if o.name == typename.name:
                        this.removeEventListener(o.type, o.listener, o.options)
                    else:
                        on[j] = o
                else:
                    on[j] = o
            on.length = len(on)
            del this.__on
        return anon

    def onAdd(self, typename, value, options):
        # TODO - write this as python
        # return function() {
        #     var on = this.__on, o, listener = contextListener(value);
        #     if (on) for (var j = 0, m = on.length; j < m; ++j) {
        #     if ((o = on[j]).type === typename.type && o.name === typename.name) {
        #         this.removeEventListener(o.type, o.listener, o.options);
        #         this.addEventListener(o.type, o.listener = listener, o.options = options);
        #         o.value = value;
        #         return;
        #     }
        #     }
        #     this.addEventListener(typename.type, listener, options);
        #     o = {type: typename.type, name: typename.name, value: value, listener: listener, options: options};
        #     if (!on) this.__on = [o];
        #     else on.push(o);
        # }
        def anon(this):
            on = this.__on
            if on:
                for j in range(0, len(on)):
                    o = on[j]
                    if o.type == typename.type and o.name == typename.name:
                        this.removeEventListener(o.type, o.listener, o.options)
                        this.addEventListener(o.type, o.listener, o.options)
                        o.value = value
                        return
                this.addEventListener(typename.type, value, options)
                o = {'type': typename.type, 'name': typename.name, 'value': value, 'listener': value, 'options': options}
                if not on:
                    this.__on = [o]
                else:
                    on.push(o)
        return anon


    def on(self, typename, value, options, *args):
        typenames = parseTypenames(str(typename))
        i = None
        n = len(typenames)
        t = None

        # TODO - write this as python
        # if (arguments.length < 2) {
        #     var on = this.node().__on;
        #     if (on) for (var j = 0, m = on.length, o; j < m; ++j) {
        #     for (i = 0, o = on[j]; i < n; ++i) {
        #         if ((t = typenames[i]).type === o.type && t.name === o.name) {
        #         return o.value;
        #         }
        #     }
        #     }
        #     return;
        # }

        # on = value ? onAdd : onRemove;
        # for (i = 0; i < n; ++i) this.each(on(typenames[i], value, options));
        # return this;
        # }

        if arguments.length < 2:
            on = self.node().__on
            if on:
                for j in range(0, len(on)):
                    o = on[j]
                    for i in range(0, n):
                        t = typenames[i]
                        if t is not None and (o.type == t.type) and (o.name == t.name):
                            return o.value
                return
            return
        for i in range(0, n):
            o = self.node().__on[i]
            for j in range(0, n):
                t = typenames[j]
                if (o.type == t.type) and (o.name == t.name):
                    o.value = value
                    return
        self.node().addEventListener(typenames[0].type, value, options)
        o = {'type': typenames[0].type, 'name': typenames[0].name, 'value': value, 'listener': value, 'options': options}
        self.node().__on.push(o)
        return self

    # import selection_dispatch from "./dispatch.js";
    # def dispatch: selection_dispatch,
    # import defaultView from "../window.js";
    def dispatchEvent(self, node, type, params):
        window = defaultView(node)
        event = window.CustomEvent
        # TODO - write this as python
        # if (typeof event === "function") {
        #     event = new event(type, params);
        # } else {
        #     event = window.document.createEvent("Event");
        #     if (params) event.initEvent(type, params.bubbles, params.cancelable), event.detail = params.detail;
        #     else event.initEvent(type, false, false);
        # }
        if callable(event):
            event = event(type, params)
        else:
            event = event.createEvent("Event")
            if params:
                event.initEvent(type, params.bubbles, params.cancelable)
            else:
                event.initEvent(type, False, False)
        node.dispatchEvent(event)
        return node

    def dispatchConstant(self, type, params):
        return lambda this: dispatchEvent(this, type, params)

    def dispatchFunction(self, type, params, *args):
        return lambda this: dispatchEvent(this, type, Object(params).apply(this, args))

    def dispatch(self, type, params):
        # return this.each((typeof params === "function"
        #     ? dispatchFunction
        #     : dispatchConstant)(type, params))
        func = dispatchFunction if callable(params) else dispatchConstant
        return this.each(func(type, params))

    # import selection_iterator from "./iterator.js";
    # #   [Symbol.iterator]: selection_iterator


def selection():
    return Selection([[document.documentElement]], root)


def selection_selection():
    return this


def select(selector):
    # print(selector)
    # print(document)
    from domonic.dom import document  # bring in the global document
    if isinstance(selector, str):
        return Selection([[document.querySelector(selector)]], [document.documentElement])
    else:
        return Selection([[selector]], root)


def create(name):
    return select(creator(name).call(document.documentElement))


# export {default as create} from "./create.js";
# export {default as creator} from "./creator.js";
# export {default as local} from "./local.js";
nextId = 0


def local():
    return Local


class Local():

    def __init__(self):
        self.nextId = 0  # += 1
        self._ = "@" + String(self.nextId).toString(36)

    def get(self, node):
        id = self._
        while not (id in node):
            node = node.parentNode
            if node is None:
                return
        return node[id]

    def set(self, node, value):
        node[this._] = value
        return node[self._]

    def remove(self, node):
        for i in range(0, len(node)):
            if node[i] == self._:
                a = node.remove(i)
                # del node[i]
                return a
        # return this._ in node and delete node[this._]

    def toString(self):
        return self._


# export {default as matcher} from "./matcher.js";
def matcher(selector):
    return lambda this: this.matches(selector)


def childMatcher(selector):
    return lambda node: node.matches(selector)


# export {default as pointer} from "./pointer.js";
# import sourceEvent from "./sourceEvent.js";
def sourceEvent(event):
    sourceEvent = event.sourceEvent
    while sourceEvent is not None:
        event = sourceEvent
        sourceEvent = event.sourceEvent
    return event


def pointer(event, node):
    event = sourceEvent(event)
    if node == None:
        node = event.currentTarget
    if node:
        svg = node.ownerSVGElement or node
    if svg.createSVGPoint:
        point = svg.createSVGPoint()
        point.x = event.clientX
        point.y = event.clientY
        point = point.matrixTransform(node.getScreenCTM().inverse())
        return [point.x, point.y]

    if (node.getBoundingClientRect):
        rect = node.getBoundingClientRect()
        return [event.clientX - rect.left - node.clientLeft, event.clientY - rect.top - node.clientTop]

    return [event.pageX, event.pageY]


def pointers(events, node):
    if events.target is not None:  # i.e., instanceof Event, not TouchList or iterable
        events = sourceEvent(events)
    if node == None:
        node = events.currentTarget
    events = events.touches or [events]
    return Array.from_(events, lambda event: pointer(event, node))


def empty():
    return []


def selectAll(selector):
    from domonic.dom import document  # bring in the global document

    # print(document)
    if isinstance(selector, str):
        return Selection([document.querySelectorAll(selector)], [document.documentElement])
        # return Selection([document.getElementsBySelector(selector, document)], [document.documentElement])
    else:
        return Selection([array(selector)], root)


def selectorAll(selector):
    return empty if selector == None else lambda: document.querySelectorAll(selector)



"""
    domonic.d3.polygon
    ====================================

"""

from domonic.javascript import Array, Math


def polygonArea(polygon):
    i = -1
    n = len(polygon)
    a = 0
    b = polygon[n - 1]
    area = 0

    while i < n:
        a = b
        b = polygon[i]
        area += a[1] * b[0] - a[0] * b[1]
        i+=1

    return area / 2


def polygonCentroid(polygon):
    i = -1
    n = len(polygon)
    x = 0
    y = 0
    a = 0
    b = polygon[n - 1]
    c = 0
    k = 0
    while i < n:
        a = b
        b = polygon[i]
        c = a[0] * b[1] - b[0] * a[1]
        k += c
        x += (a[0] + b[0]) * c
        y += (a[1] + b[1]) * c
        i+=1

    k *= 3
    return [x / k, y / k]

def cross(a, b, c):
    """[Returns the 2D cross product of AB and AC vectors, i.e., the z-component of
        the 3D cross product in a quadrant I Cartesian coordinate system (+x is
        right, +y is up). Returns a positive value if ABC is counter-clockwise,
        negative if clockwise, and zero if the points are collinear.]

    Args:
        a ([type]): [description]
        b ([type]): [description]
        c ([type]): [description]

    Returns:
        [type]: [description]
    """
    return (b[0] - a[0]) * (c[1] - a[1]) - (b[1] - a[1]) * (c[0] - a[0])

def polygonHull(points):

    n = len(points)
    if n < 3:
        return None

    # Convert list to tuples and remove duplicates
    points = [t for t in (set(tuple(i) for i in points))]

    # Sort lexicographically 
    points = sorted(points)

    # Build hulls according to Andrew's monotone chain algorithm
    lowerHull = []
    for p in points:
        while len(lowerHull) >= 2 and cross(lowerHull[-2], lowerHull[-1], p) <=0 :
            lowerHull.pop()
        lowerHull.append(p)

    upperHull = []
    for p in reversed(points):
        while len(upperHull) >= 2 and cross(upperHull[-2], upperHull[-1], p) <= 0:
            upperHull.pop()
        upperHull.append(p)
    
    # Build polygon hull from upper/lower hulls and return to list form
    hull = sorted([list(i) for i in lowerHull[:-1] + upperHull[:-1]])
    
    return hull


def polygonContains(polygon, point):
    n = len(polygon)
    p = polygon[n - 1]
    x = point[0]
    y = point[1]
    x0 = p[0]
    y0 = p[1]
    x1 = 0
    y1 = 0
    inside = False
    for i in range(0, n):
        p = polygon[i]
        x1 = p[0]
        y1 = p[1]
        if (((y1 > y) != (y0 > y)) and (x < (x0 - x1) * (y - y1) / (y0 - y1) + x1)):
            inside = not inside
        x0 = x1
        y0 = y1
    return inside


def polygonLength(polygon):
    i = -1
    n = len(polygon)
    b = polygon[n - 1]
    xa = 0
    ya = 0
    xb = b[0]
    yb = b[1]
    perimeter = 0
    while i < n:
        xa = xb
        ya = yb
        b = polygon[i]
        xb = b[0]
        yb = b[1]
        xa -= xb
        ya -= yb
        perimeter += Math.hypot(xa, ya)
        i += 1
    return perimeter


"""
    domonic.d3.geom.shape
    ====================================

"""

from domonic.geom.vec2 import vec2
from domonic.svg import *


class vertex(vec2):
    """
    A vertex is a point in the shape.
    """

    # def __init__(self, x=0, y=0):
    #     self.x = x
    #     self.y = y

    def move_to(self, position):
        x, y = position
        self.x = x
        self.y = y

    def move_by(self, dx, dy):
        self.x += dx
        self.y += dy

    # def __str__(self):
    #     return '{} {} {}'.format(self.x, self.y, self.color)

    # def __repr__(self):
    #     return '{} {} {}'.format(self.x, self.y, self.color)


Point = vertex


class Shape(Point):
    def __init__(self, x=0, y=0, color="red", vertices=[]):
        super().__init__(x, y)

        if isinstance(vertices, list):
            self.vertices = vertices

        self.color = color
        self.scale = 1
        self.rotation = 0
        # self.pivot = (0, 0)
        self.id = None
        self.name = None
        self.style = None
        self.locked = False
        self.visible = True
        self.selected = False
        self.clipMask = None
        self.data = None
        self.blendMode = None
        self.opacity = 1.0
        self.strokeColor = None
        self.strokeWidth = 1
        # self.strokeCap = 'butt'
        # self.strokeJoin = 'miter'
        # self.dashOffset = 0
        # self.strokeScaling = 1.0
        # self.dashArray = None
        # self.miterLimit = None
        # self.fillColor = None
        self.shadowColor = None
        self.shadowBlur = None
        self.shadowOffset = None
        self.shadowOpacity = None
        # self.vertcies = []
        # self.width = None
        # self.height = None

    @property
    def position(self):
        return vec2(self.x, self.y)

    @position.setter
    def position(self, pos):
        self.x = pos.x
        self.y = pos.y

    @property
    def width(self):
        """determine the width of the shape"""
        return (
            max(self.vertices, key=lambda v: v.x).x
            - min(self.vertices, key=lambda v: v.x).x
        )

    @property
    def height(self):
        """determine the height of the shape"""
        return (
            max(self.vertices, key=lambda v: v.y).y
            - min(self.vertices, key=lambda v: v.y).y
        )

    def rotate(self, angle):
        """rotate the shape"""
        self.rotation += angle

    def draw(self, svg):
        """draw the shape"""
        if self.visible:
            svg.add(
                svg.polyline(
                    self.vertices,
                    id=self.id,
                    class_=self.name,
                    style=self.style,
                    fill=self.color,
                    stroke=self.strokeColor,
                    stroke_width=self.strokeWidth,
                    opacity=self.opacity,
                    fill_opacity=self.opacity,
                    fill_rule="evenodd",
                )
            )

    def __len__(self):
        return len(self.vertices)

    def __getitem__(self, index):
        return self.vertices[index]

    def __setitem__(self, index, value):
        self.vertices[index] = value

    def __delitem__(self, index):
        del self.vertices[index]

    def __iter__(self):
        return iter(self.vertices)

    def __reversed__(self):
        return reversed(self.vertices)

    def __contains__(self, item):
        return item in self.vertices

    def __add__(self, other):
        """add two shapes"""
        if isinstance(other, Shape):
            return self.vertices + other.vertices

    # def matrix
    # def clear
    # def pivot
    # def bounds

    # def rasterize(self, rasterizer):
    #     """ rasterize the shape """
    #     pass

    # def contains(self, point):
    #     """ check if the point is inside the shape """
    #     pass

    # def isInside(self, rect):
    #     """ check if the rect is inside the shape """
    #     pass

    # def intersects(item):
    #     """ check if the shape intersects with another """
    #     pass

    # @property
    # def x(self):
    #     return self.__x

    # @x.setter
    # def x(self, x):
    #     self.__x = x

    # @property
    # def y(self):
    #     return self.__y

    # @y.setter
    # def y(self, y):
    #     self.__y = y


class Line(Shape):
    def __init__(self, p1, p2, color=None, *args):
        super().__init__(color)
        # if isinstance(p1, vec2):
        self.p1 = p1
        self.p2 = p2
        self.x = self.p1[0] - self.p2[0]
        self.y = self.p1[1] - self.p2[1]

    def __str__(self):
        return f"Line({self.p1}, {self.p2})"

    def __eq__(self, other):
        return self.p1 == other.p1 and self.p2 == other.p2

    def __ne__(self, other):
        return self.p1 != other.p1 or self.p2 != other.p2

    def __hash__(self):
        return hash((self.p1, self.p2))

    def __iter__(self):
        return iter((self.p1, self.p2))

    # def __len__(self):
    #     return 2

    def __getitem__(self, index):
        return (self.p1, self.p2)[index]

    def __setitem__(self, index, value):
        if index == 0:
            self.p1 = value

    def __delitem__(self, index):
        if index == 0:
            self.p1 = None

    def __contains__(self, item):
        return item in (self.p1, self.p2)

    def __add__(self, other):
        return Line(self.p1, other.p2)

    def __sub__(self, other):  # self - other
        return Line(self.p1, other.p1)

    def __mul__(self, other):
        return Line(self.p1, other.p2)

    def __rmul__(self, other):
        return Line(self.p1, other.p2)

    def __neg__(self):
        return Line(self.p1, self.p2)

    def __pos__(self):
        return Line(self.p1, self.p2)

    def __abs__(self):
        return Line(self.p1, self.p2)

    # def __bool__(self):
    #     return bool(self.p1 and self.p2)

    # def __nonzero__(self):
    #     return bool(self.p1 and self.p2)

    # def __call__(self, p):
    #     return Line(self.p1, p)

    # def __getattr__(self, name):
    #     if name == 'p1':
    #         return self.p1
    #     elif name == 'p2':
    #         return self.p2
    #     else:
    #         raise AttributeError(name)

    # def __setattr__(self, name, value):
    #     if name == 'p1':
    #         self.p1 = value

    # def __delattr__(self, name):
    #     if name == 'p1':
    #         self.p1 = None
    #     elif name == 'p2':
    #         self.p2 = None
    #     else:
    #         raise AttributeError(name)

    # def __getinitargs__(self):
    #     return (self.p1, self.p2)

    # def __setstate__(self, state):
    #     self.p1 = state['p1']
    #     self.p2 = state['p2']

    def __getstate__(self):
        return {"p1": self.p1, "p2": self.p2}

    def __reduce__(self):
        return (Line, (self.p1, self.p2))

    def __reduce_ex__(self, protocol):
        return self.__reduce__()

    def __copy__(self):
        return Line(self.p1, self.p2)

    def __deepcopy__(self, memo):
        return Line(self.p1, self.p2)

    def __bool__(self):
        return bool(self.p1 and self.p2)

    def __nonzero__(self):
        return bool(self.p1 and self.p2)


class Plane:
    def __init__(self, normal, distance, color=None, *args):
        """[a plane is defined by its normal vector and a distance from the origin]

        Args:
            normal ([vec2]): [a vector representing the normal vector of the plane]
            distance ([type]): [a scalar representing the distance from the origin]
            color ([vec3], optional): [the color of the plane]
        """
        self.normal = normal
        self.distance = distance
        self.color = color

    def __str__(self):
        return f"Plane({self.normal}, {self.distance})"

    def __eq__(self, other):
        return self.normal == other.normal and self.distance == other.distance

    # def bisectors(self):
    #     return self.normal.cross(self.distance)

    # def with_distance(self, distance):
    #     return Plane(self.normal, distance)

    # def with_normal(self, normal):
    #     return Plane(normal, self.distance)


class Rect(Shape):
    def __init__(self, x=0, y=0, width=1, height=1, color=None):
        super().__init__(color)
        self.x = x
        self.y = y
        self.width = width
        self.height = height

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, w):
        self._width = w

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, w):
        self._height = w

    def get_bottom_right(self):
        d = self.x + self.width
        t = self.y + self.height
        return (d, t)

    def get_bottom_left(self):
        b = self.get_bottom_right()
        l = super().get_bottom_right()
        return (b[0], l[0])

    def get_top_left(self):
        t = self.x
        b = super().get_bottom_right()
        return (t, b[1])

    def get_top_right(self):
        t = self.x + self.width
        b = super().get_bottom_right()
        return (t, b[1])

    def get_center(self):
        c = self.x + self.width / 2
        t = self.y + self.height / 2
        return (c, t)

    def get_center_x(self):
        c = self.x + self.width / 2
        return c

    def get_center_y(self):
        c = self.y + self.height / 2
        return c

    def get_left(self):
        return self.x

    def get_right(self):
        return self.x + self.width

    def get_top(self):
        return self.y

    def get_bottom(self):
        return self.y + self.height

    def get_width(self):
        return self.width

    def get_height(self):
        return self.height

    def __str__(self):
        return f"Rect({self.x}, {self.y}, {self.width}, {self.height})"

    def __eq__(self, other):
        return (
            self.x == other.x
            and self.y == other.y
            and self.width == other.width
            and self.height == other.height
        )

    def __ne__(self, other):
        return (
            self.x != other.x
            or self.y != other.y
            or self.width != other.width
            or self.height != other.height
        )

    def __hash__(self):
        return hash((self.x, self.y, self.width, self.height))

    def __iter__(self):
        return iter((self.x, self.y, self.width, self.height))

    def __len__(self):
        return 4

    def __getitem__(self, index):
        return (self.x, self.y, self.width, self.height)[index]

    def __setitem__(self, index, value):
        if index == 0:
            self.x = value
        elif index == 1:
            self.y = value
        elif index == 2:
            self.width = value
        elif index == 3:
            self.height = value

    def __add__(self, other):
        return Rect(
            self.x + other.x,
            self.y + other.y,
            self.width + other.width,
            self.height + other.height,
        )

    def __sub__(self, other):
        return Rect(
            self.x - other.x,
            self.y - other.y,
            self.width - other.width,
            self.height - other.height,
        )

    def __mul__(self, other):
        return Rect(
            self.x * other, self.y * other, self.width * other, self.height * other
        )

    def __truediv__(self, other):
        return Rect(
            self.x / other, self.y / other, self.width / other, self.height / other
        )

    def __floordiv__(self, other):
        return Rect(
            self.x // other, self.y // other, self.width // other, self.height // other
        )

    def __iadd__(self, other):
        self.x += other.x
        self.y += other.y
        self.width += other.width
        self.height += other.height
        return self

    def __isub__(self, other):
        self.x -= other.x
        self.y -= other.y
        self.width -= other.width
        self.height -= other.height
        return self

    def __imul__(self, other):
        self.x *= other
        self.y *= other
        self.width *= other
        self.height *= other
        return self

    def __idiv__(self, other):
        self.x /= other
        self.y /= other
        self.width /= other
        self.height /= other
        return self

    def __ifloordiv__(self, other):
        self.x //= other
        self.y //= other
        self.width //= other
        self.height //= other
        return self

    # def __getstate__(self):
    # return {'x': self.x, 'y': self.y, 'width': self.width, 'height': self.height}

    # def __setstate__(self, state):
    #     self.x = state['x']
    #     self.y = state['y']
    #     self.width = state['width']
    #     self.height = state['height']

    def __reduce__(self):
        return (Rect, (self.x, self.y, self.width, self.height))

    def __copy__(self):
        return Rect(self.x, self.y, self.width, self.height)

    def __deepcopy__(self, memo):
        return Rect(self.x, self.y, self.width, self.height)

    def __contains__(self, other):
        return (
            self.x <= other.x
            and self.y <= other.y
            and self.x + self.width >= other.x + other.width
            and self.y + self.height >= other.y + other.height
        )

    def __lt__(self, other):
        return (
            self.x < other.x
            and self.y < other.y
            and self.width < other.width
            and self.height < other.height
        )

    def __le__(self, other):
        return (
            self.x <= other.x
            and self.y <= other.y
            and self.width <= other.width
            and self.height <= other.height
        )

    def __gt__(self, other):
        return (
            self.x > other.x
            and self.y > other.y
            and self.width > other.width
            and self.height > other.height
        )

    def __ge__(self, other):
        return (
            self.x >= other.x
            and self.y >= other.y
            and self.width >= other.width
            and self.height >= other.height
        )

    # def __eq__(self, other):
    #     return self.x == other.x and self.y == other.y and self.width == other.width and self.height == other.height

    # def __ne__(self, other):
    #     return self.x != other.x or self.y != other.y or self.width != other.width or self.height != other.height

    # def __hash__(self):
    #     return hash((self.x, self.y, self.width, self.height))


class Square(Rect):
    def __init__(self, x=0, y=0, size=1.0, color=None):
        super().__init__(x, y, size, size, color)

    def get_size(self):
        return self.width

    def set_size(self, size):
        self.width = size
        self.height = size

    def __str__(self):
        return f"Square({self.x}, {self.y}, {self.width})"

    # def __getstate__(self):
    #     return {'x': self.x, 'y': self.y, 'size': self.width, 'color': self.color}

    # def __setstate__(self, state):
    #     self.x = state['x']
    #     self.y = state['y']
    #     self.width = state['size']
    #     self.height = state['size']
    #     self.color = state['color']

    # def __reduce__(self):
    #     return (Square, (self.x, self.y, self.width, self.color))

    def __copy__(self):
        return Square(self.x, self.y, self.width, self.color)

    def __deepcopy__(self, memo):
        return Square(self.x, self.y, self.width, self.color)

    def __contains__(self, other):
        return (
            self.x <= other.x
            and self.y <= other.y
            and self.x + self.width >= other.x + other.width
            and self.y + self.height >= other.y + other.height
        )

    def get_vertices(self, x=None, y=None):
        if x is None:
            x = self.x
        if y is None:
            y = self.y
        return [
            (x, y),
            (x + self.width, y),
            (x + self.width, y + self.height),
            (x, y + self.height),
        ]

    def get_vertices_list(self):
        return [
            (self.x, self.y),
            (self.x + self.width, self.y),
            (self.x + self.width, self.y + self.height),
            (self.x, self.y + self.height),
        ]

    def svg(self, x=None, y=None):
        if x is None:
            x = self.x
        if y is None:
            y = self.y
        return f'<rect x="{x}" y="{y}" width="{self.width}" height="{self.height}" fill="{self.color}"/>'

    # def draw_to_canvas(self, canvas):
    #     canvas.draw_rect(self.x, self.y, self.width, self.height, self.color)


class Polygon(Shape):
    def __init__(self, points):
        super().__init__()
        self.points = points

    def get_points(self):
        return self.points

    def set_points(self, points):
        self.points = points

    def __str__(self):
        return f"Polygon({(self.points)})"

class Polyline(Shape):
    def __init__(self, points):
        super().__init__()
        self.points = points

    def get_points(self):
        return self.points

    def set_points(self, points):
        self.points = points

    def __str__(self):
        return f"Polyline({(self.points)})"

    def __getstate__(self):
        return {"points": self.points}

    def __setstate__(self, state):
        self.points = state["points"]

    def __reduce__(self):
        return (Polyline, (self.points,))

    def __copy__(self):
        return Polyline(self.points)

    def __deepcopy__(self, memo):
        return Polyline(self.points)

    def __contains__(self, other):
        return other in self.points

    def __len__(self):
        return len(self.points)

    def __getitem__(self, key):
        return self.points[key]

    def __setitem__(self, key, value):
        self.points[key] = value

    def __iter__(self):
        return iter(self.points)

    def __add__(self, other):
        return Polyline(self.points + other.points)

    def __sub__(self, other):
        return Polyline(self.points - other.points)

    def __mul__(self, other):
        return Polyline(self.points * other)

    def __rmul__(self, other):
        return Polyline(other * self.points)

    def __floordiv__(self, other):
        return Polyline(self.points // other)

    def __mod__(self, other):
        return Polyline(self.points % other)

    def __divmod__(self, other):
        return Polyline(divmod(self.points, other))

    def __pow__(self, other):
        return Polyline(self.points ** other)

    def __lshift__(self, other):
        return Polyline(self.points << other)

    def __rshift__(self, other):
        return Polyline(self.points >> other)

    def __and__(self, other):
        return Polyline(self.points & other)

    def __xor__(self, other):
        return Polyline(self.points ^ other)

    def __or__(self, other):
        return Polyline(self.points | other)

    def __iadd__(self, other):
        self.points += other.points
        return self

    def __isub__(self, other):
        self.points -= other.points
        return self

    def __imul__(self, other):
        self.points *= other
        return self

    def __imod__(self, other):
        self.points %= other
        return self

    def __idivmod__(self, other):
        self.points //= other
        return self

    def __ipow__(self, other):
        self.points **= other

    def __ilshift__(self, other):
        self.points <<= other

    def __irshift__(self, other):
        self.points >>= other
        return self

    def __iand__(self, other):
        self.points &= other

    def __ixor__(self, other):
        self.points ^= other

    def __ior__(self, other):
        self.points |= other
        return self

    def __neg__(self):
        return Polyline(-self.points)

    def __pos__(self):
        return Polyline(self.points)

    def __abs__(self):
        return Polyline(abs(self.points))

    def __invert__(self):
        return Polyline(~self.points)

    # def __complex__(self):
    #     return complex(self.points)

    def __int__(self):
        return int(self.points)

    def __float__(self):
        return float(self.points)

    def __round__(self):
        return round(self.points)

    # def __reversed__(self):
    #     return Polyline(reversed(self.points))


# class Quad(Shape):
#     def __init__(self, left=Rect(), right=Rect(), front=Rect(), back=Rect()):
#         super().__init__(color='blue')
#         self.left, self.right, self.front, self.back = left, right, front, back

# class Triangle(Shape):
#     def __init__(self):
#         super().__init__(self.color='blue')


# class Wave():

#     def __init__(self, color=None, *args):
#         self.color = color


from domonic.geom.shape.shapes import Circle, Oval

# from domonic.geom.shape.circle import Elipse



"""
    domonic.webapi.fetch
    ====================================
    https://developer.mozilla.org/en-US/docs/Web/API/Fetch_API
"""

# import array
# import datetime
# import gc
# import json
# import math
# import multiprocessing
# import os
# import random
# import re
# import signal
# import struct
# import sys
import threading
from multiprocessing.pool import ThreadPool as Pool

from domonic.window import window
from domonic.javascript import Promise
# TODO - untested. moving these over from javascript module
# TODO - check if promise also needs to come to this package


class FetchedSet:  # not a promise
    def __init__(self, *args, **kwargs):
        self.results = []

    def __getitem__(self, index):
        return self.results[index]

    def oncomplete(self, func):  # runs once all results are back
        func(self.results)
        return

    # def __call__(self, func):
    #     self.results.append(func)

# @staticmethod
def fetch(url: str, **kwargs):
    # undocumented - warning. use at own risk
    # note - kinda pointless atm. just use requests directly and you wont have to muck about with a Promise
    if type(url) is not str:
        raise ValueError('fetch takes a single url string. use fetch_set, fetch_threaded or fetch_pooled')
    f = Promise()
    r = window._do_request(url, f, *kwargs)
    return f.resolve(r)

# @staticmethod
def fetch_set(urls: list, callback_function=None, error_handler=None, **kwargs):
    # undocumented - warning. use at own risk
    # note - still blocks. just gets all before continuing
    # problems - all urls can only have 1 associated callback, error and set of kwargs
    if type(urls) is str:
        urls = [urls]  # leniency
    f = FetchedSet()
    for url in urls:
        r = window.fetch(url, **kwargs).then(callback_function)
        f.results.append(r.data)
    return f

# @staticmethod
def fetch_threaded(urls: list, callback_function=None, error_handler=None, **kwargs):
    # undocumented - warning. use at own risk
    # note - still blocks. just gets all before continuing using threads
    # problems - all urls can only have 1 associated callback, error and set of kwargs
    if type(urls) is str:
        urls = [urls]  # leniency
    f = FetchedSet()
    jobs = []
    for url in urls:
        thread = threading.Thread(target=window._do_request(url, f, **kwargs))
        thread.setDaemon(True)
        jobs.append(thread)
    map(lambda j: j.start(), jobs)
    map(lambda j: j.join(), jobs)
    # f = FetchedSet()
    return f

# @staticmethod
def fetch_pooled(urls: list, callback_function=None, error_handler=None, **kwargs):
    # undocumented - warning. use at own risk
    # note - still blocks. just gets all before continuing using a pool
    # problems - all urls can only have 1 associated callback, error and set of kwargs
    if type(urls) is str:
        urls = [urls]  # leniency
    f = FetchedSet()

    def _do_request_wrapper(obj):
        url = obj['url']
        f = obj['f']
        kwargs = obj['k']
        kwargs['callback_function'] = obj['c']
        kwargs['error_handler'] = obj['e']
        window._do_request(url, f, **kwargs)

    jobs = []
    p = Pool()
    urls = [{'url': url, 'f': f, 'c': callback_function, 'e': error_handler, 'k': kwargs} for url in urls]
    results = p.map(_do_request_wrapper, urls)
    p.close()
    p.join()
    return f

# def fetch_aysnc( urls: list, options={}, type="async" ):
    # TODO - a version using async/await


class Headers():

    def __init__(self, all=None):
        # https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/getAllResponseHeaders#Example
        self.headers = {}  #Object.create(None)
        arr = all.split("\r\n")
        for each in arr:
            line = arr
            parts = line.split(": ")
            name = parts.shift()
            value = parts.join(": ")
            self.headers[name.lower()] = value

    def get(self, name: str):
        return self.headers[name.lower()]

    def set(self, name: str, value: str):
        self.headers[name.lower()] = value

    def has(self, name: str):
        return name in self.headers

    def keys(self):
        return self.headers.keys()

    def values(self):
        return self.headers.values()

    def entries(self):
        return self.headers.entries()

    def delete(self, name):
        del self.headers[name]

    def forEach(self, callback, thisArg=None):
        for name, value in self.headers.items():
            callback(value, name, self)

    def map(self, callback, thisArg=None):
        return [callback(value, name, self) for name, value in self.headers.items()]

    def filter(self, callback, thisArg=None):
        return [callback(value, name, self) for name, value in self.headers.items() if callback(value, name, self)]

    def reduce(self, callback, initialValue):
        return [callback(initialValue, value, name, self) for name, value in self.headers.items()]

    def toString(self):
        return str(self.headers)

    def toObject(self):
        return self.headers

    def toJSON(self):
        return self.headers

    def __str__(self):
        return str(self.headers)

    def __repr__(self):
        return str(self.headers)

    def __iter__(self):
        return self.headers.__iter__()

    def __next__(self):
        return self.headers.__next__()

    def __getitem__(self, key):
        return self.headers[key]

    def __setitem__(self, key, value):
        self.headers[key] = value

    def __delitem__(self, key):
        del self.headers[key]

    def __contains__(self, key):
        return key in self.headers


class Response():
    def __init__(self, url=None, status=None, statusText=None, headers=None, body=None):
        self.url = url
        self.status = status
        self.statusText = statusText
        self.headers = headers
        self.body = body

    def arrayBuffer(self):
        return self.body

    def blob(self):
        return self.body

    def formData(self):
        return self.body

    def json(self):
        return self.body

    def text(self):
        return self.body

    def clone(self):
        return Response(self.url, self.status, self.statusText, self.headers, self.body)

    def __str__(self):
        return str(self.body)

    def __repr__(self):
        return str(self.body)

    def __iter__(self):
        return self.body.__iter__()

    def __next__(self):
        return self.body.__next__()

    def __getitem__(self, key):
        return self.body[key]

    def __setitem__(self, key, value):
        self.body[key] = value

    def __delitem__(self, key):
        del self.body[key]

    def __contains__(self, key):
        return key in self.body


class Request():
    def __init__(self, url=None, method=None, headers=None, body=None, mode=None, credentials=None, cache=None):
        self.url = url
        self.method = method
        self.headers = headers
        self.body = body
        self.mode = mode
        self.credentials = credentials
        self.cache = cache

    def clone(self):
        return Request(self.url, self.method, self.headers, self.body, self.mode, self.credentials, self.cache)

    def arrayBuffer(self):
        return self.body



"""
    domonic.parsers
    ====================================

    A place for parsers and utils for doings so.

    These methods operate strings not on pyml objects. For dom manipulation use the dom.

    WARNING> dont use this class. still in development/idea phase. Teasing util methods out from the in-place parser

"""

import re

from domonic.html import *
from domonic.xml.sitemap import *


def create_element(name='custom_tag', *args, **kwargs):
    '''
    NOTE - USED BY THE HACKED EXPAT PARSER TO GET VALID DOCUMENT NODES FROM ANY KNOWN SET
    '''
    from domonic.html import html_tags
    if name in html_tags:
        return globals()[name]()
    from domonic.xml.sitemap import sitemap_tags
    if name in sitemap_tags:
        return globals()[name]()

    from domonic.html import Element, tag  # , tag_init
    custom_tag = type('custom_tag', (tag, Element), {'name': name})   #, '__init__': tag_init})
    new_tag = custom_tag(*args, **kwargs)
    new_tag.name = name
    return new_tag


def remove_tags(html_str: str, tags):
    """
    removes a list of tags and their content from the html
    """
    if isinstance(tags, str):
        tags = [tags]

    if isinstance(tags, list):
        for tag in tags:

            if tag == 'js' or tag == 'javascript':
                scripts = re.compile(r'<(script).*?</\1>(?s)')
                html_str = scripts.sub('', html_str)

            if tag == 'css':
                css = re.compile(r'<(style).*?</\1>(?s)')
                html_str = css.sub('', html_str)

            if 'comment' in tag or tag == '#' or tag == '//':
                comments = re.compile(r'<!--(.|\s)*?-->')
                html_str = comments.sub('', html_str)

            # tag = re.compile(r'<(style).*?</\1>(?s)')
            # html = tag.sub('', html)
    return html_str


def remove_extra_whitespace(html_str: str):
    """
    only allow single spaces and tabs
    """
    html_str = re.sub(r'\s+', ' ', html_str)
    html_str = re.sub(r'\t', ' ', html_str)
    return html_str
    # remove abnormal spacing between tag attributes (TODO- maybe 2 spaces is valid somewhere?)
    # page = page.replace('   ', ' ')
    # page = page.replace('  ', ' ')
    # page = page.replace('  ', ' ')


def remove_doctype(html_str: str):
    """
    remove the doctype from the html_str
    """
    doctype = re.compile(r'<!DOCTYPE.*?>', re.IGNORECASE)
    html_str = doctype.sub('', html_str)
    return html_str
    # print('parsing parsing parsing!!')

    page = ''.join(page.split('<!DOCTYPE HTML>'))
    page = ''.join(page.split('<!DOCTYPE html>'))
    page = ''.join(page.split('<!doctype html>'))
    page = ''.join(page.split('<!doctype HTML>'))

    page = ''.join(page.split('<?xml version="1.0" encoding="utf-8"?>'))
    page = ''.join(page.split('<?xml version="1.0" encoding="utf-8" ?>'))
    page = ''.join(page.split('<?xml version="1.0" encoding="UTF-8" ?>'))

    page = ''.join(page.split('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">'))
    # page = ''.join(page.split('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">'))
    # page = ''.join(page.split('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Frameset//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd">'))
    # page = ''.join(page.split('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">'))
    # page = ''.join(page.split('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">'))
    # page = ''.join(page.split('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Frameset//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd">'))
    # page = ''.join(page.split('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">'))
    # page = ''.join(page.split('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">'))
    # page = ''.join(page.split('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Frameset//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd">'))
    # page = ''.join(page.split('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">'))
    # page = ''.join(page.split('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">'))


def remove_xml_header(html_str: str):
    """
    remove the xml header from the html_str
    """
    header = re.compile(r'<\?xml.*?\?>', re.IGNORECASE)
    html_str = header.sub('', html_str)
    return html_str


def remove_html_tags(html_str: str):
    """
    remove all html tags from the html_str
    """
    # remove all tags
    page = re.compile(r'<.*?>', re.IGNORECASE)
    html_str = page.sub('', html_str)
    return html_str


def remove_html_tag_by_name(html_str: str, tag: str):
    """
    remove all html tags with the given name from the html_str
    """
    # remove all tags
    page = re.compile(r'<.*?{}.*?>'.format(tag), re.IGNORECASE)
    html_str = page.sub('', html_str)
    return html_str


# def remove_all_tags(html: str):
#     """
#     remove tags
#     """
#     page = re.sub(r'<[^>]*>', '', page)

def remove_content_between_brackets(html_str: str):
    """
    remove content between brackets
    """
    page = re.sub(r'\[[^\]]*\]', '', html_str)
    return page


def remove_content_between_parenthesis(html_str: str):
    """
    remove content between parenthesis
    """
    page = re.sub(r'\([^\)]*\)', '', html_str)
    return page


def remove_content_between_braces(html_str: str):
    """
    remove content between braces
    """
    page = re.sub(r'\{[^\}]*\}', '', html_str)
    return page

# def remove_consecutive_spaces(html_str: str):
#     """
#     remove consecutive spaces
#     """
#     page = re.sub(r'\s+', ' ', page)


def remove_whitespace(html_str: str):
    """
    remove whitespace
    """
    html_str = re.sub(r'\s+', ' ', html_str)
    return html_str


def remove_newlines(html_str: str):
    """
    remove newlines
    """
    html_str = re.sub(r'\n', '', html_str)
    return html_str


def remove_tabs(html_str: str):
    """
    remove all tabs from the html_str
    """
    html_str = re.sub(r'\t', '', html_str)
    return html_str


def replace_special_quotes(html_str: str):
    """
    replace special quotes with html entities
    """
    # special quotes
    html_str = html_str.replace('“', '&ldquo;')
    html_str = html_str.replace('”', '&rdquo;')
    html_str = html_str.replace('’', '&rsquo;')
    html_str = html_str.replace('‘', '&lsquo;')
    html_str = html_str.replace('„', '&sbquo;')
    html_str = html_str.replace('‚', '&obquo;')
    html_str = html_str.replace('‹', '&usbquo;')
    html_str = html_str.replace('›', '&ensquo;')
    return html_str


def replace_special_chars(html_str: str):
    """
    replace special characters with html entities
    """
    # special chars
    html_str = html_str.replace('&', '&amp;')
    html_str = html_str.replace('<', '&lt;')
    html_str = html_str.replace('>', '&gt;')
    html_str = html_str.replace('"', '&quot;')
    html_str = html_str.replace("'", '&#39;')
    return html_str


# def remove_bom(html: str):
    # page = page.replace('\ufeff', '')

def replace_punctuation(html_str: str):
    """
    replace punctuation with html entities
    """
    # special chars
    html_str = html_str.replace('.', '&#46;')
    html_str = html_str.replace(',', '&#44;')
    html_str = html_str.replace('!', '&#33;')
    html_str = html_str.replace('?', '&#63;')
    html_str = html_str.replace('(', '&#40;')
    html_str = html_str.replace(')', '&#41;')
    html_str = html_str.replace('[', '&#91;')
    html_str = html_str.replace(']', '&#93;')
    html_str = html_str.replace('{', '&#123;')
    html_str = html_str.replace('}', '&#125;')
    html_str = html_str.replace('<', '&lt;')
    html_str = html_str.replace('>', '&gt;')
    html_str = html_str.replace('"', '&quot;')
    html_str = html_str.replace("'", '&#39;')
    return html_str


# def add_newlines(html: str):
    # """
    # add newlines
    # """
    # page = page.replace('\n', '<br>')

def replace_newlines(html_str: str):
    """
    remove newlines
    """
    html_str = html_str.replace('<br>', '\n')
    return html_str

# def add_paragraphs(html: str):
#     """
#     add paragraphs
#     """
#     page = page.replace('\n', '<br>\n')

# def remove_paragraphs(html: str):
#     """
#     remove paragraphs
#     """
#     page = page.replace('<br>\n', '\n')


def clean_junk(page):
    """[clears any typically invalid runs of chars that may exist in pyml.
        to be used at the end of all cleaning functions before an evaluation.
        should not apply to content.]

    Args:
        page ([str]): [pyml string]

    Returns:
        [str]: [pyml string with garbled chars removed]
    """
    page = page.replace('",","', '","')
    page = page.replace('",",', '",')
    page = page.replace('", ",', '",')
    page = page.replace('", \n",', '",')
    page = page.replace('",\n ",', '",')
    page = page.replace('","\n)', '"\n)')
    page = page.replace('", \n, _', '",\n_')
    page = page.replace(', "\n)', '\n)')

    # page = page.replace(',",', ',') < VALID
    page = page.replace(',  ,', ',')  # < new bug. due to single attributes having big space in front for some reason
    page = page.replace(', ",', ',')
    page = page.replace(',,', ',')
    page = page.replace(', ,', ',')
    page = page.replace(',"",', ',')  # careul. new and covers up somethings else. solo attributes still not done well

    page = page.replace('( ,*', '(*')
    page = page.replace('( , *', '(*')
    page = page.replace('(,*', '(*')
    page = page.replace('(, *', '(*')
    page = page.replace('(,  *', '(*')
    page = page.replace('(", *', '(*')
    page = page.replace('(\n",', '(\n')
    page = page.replace('(\n,', '(')

    page = page.replace('),",', '),')
    page = page.replace('),\n"\n),', '),\n),')
    page = page.replace('},\n"\n),', '}\n),')

    page = page.replace('"_, _', '"_')  # when solo hyphenated custom attribute is first on a line.

    # page = page.replace('),\n",\n', '(')  # < break things but is also valid. text sentences can start with a comma
    # 2 issues. this also turns a closer into an opener. when catching a true case
    return page


# @staticmethod
def dent(pyml, use_tabs=False):
    """ [
        proper dentage for pyml
    ]
    """
    tabs_or_spaces = "    "
    if use_tabs:
        tabs_or_spaces = "\t"
    dentage = 0
    lastchar = ""
    dented = ""
    for count, char in enumerate(pyml):
        if char == "(":
            dentage += 1
        if char == ")":
            dentage -= 1
        if lastchar == "\n":  # TODO - if file doesn't have newlines already
            char = tabs_or_spaces * dentage + char
        lastchar = char
        dented += char
        if dentage < 0:
            dentage = 0
    return dented


def add_cdata_tags_to_every_node(content: str):  # TODO - just have a CDATASection class?
    """[puts a CDATA tag on every node in the document] """
    content = content.replace('<', '<![CDATA[')
    content = content.replace('>', ']]>')
    return content


def remove_cdata_tags_from_every_node(content: str):
    """[removes a CDATA tag from every node in the document] """
    content = content.replace(']]>', '>')
    content = content.replace('<![CDATA[', '<')
    return content


def add_xml_declaration_to_document(content: str):
    """[puts an XML declaration at the top of the document] """
    content = content.replace('<', '<?xml version="1.0" encoding="UTF-8" ?>\n<')
    return content


# TODO - other methods I can add to this class
# - add_xml_declaration_to_document
# - add_cdata_tags_to_every_node
# - remove_cdata_tags_from_every_node
# - remove_newlines
# - add_paragraphs
# - indent
# - create_html_from_markdown
# - create_html_from_xml
# - get_xpath_of_node


''' TODO - check earlier version getter/setters for some string methods that may be useful here
@body.setter
def body(self, content):
    """ Sets the document's body (the <body> element) """
    # self.querySelector('body')
    # tag = "body"
    # reg = f"<{tag}.*?>(.+?)</{tag}>"
    # pattern = re.compile(reg)
    # tags = re.findall(pattern,html)
    # return tags[0]
    print("TODO - setter method on body")
    return
'''



"""
    domonic.html
    ====================================

    Generate HTML using python.

"""
from domonic.dom import (Comment, Document,  # HTMLOptionsCollection,
                         DocumentType, Element, HTMLAnchorElement,
                         HTMLAreaElement, HTMLAudioElement, HTMLBaseElement,
                         HTMLBaseFontElement, HTMLBodyElement, HTMLBRElement,
                         HTMLButtonElement, HTMLCanvasElement,
                         HTMLContentElement, HTMLDataElement,
                         HTMLDataListElement, HTMLDialogElement,
                         HTMLDivElement, HTMLDListElement, HTMLDocument,
                         HTMLElement, HTMLEmbedElement, HTMLFieldSetElement,
                         HTMLFormControlsCollection, HTMLFormElement,
                         HTMLFrameSetElement, HTMLHeadElement,
                         HTMLHeadingElement, HTMLHRElement, HTMLIFrameElement,
                         HTMLImageElement, HTMLInputElement,
                         HTMLIsIndexElement, HTMLKeygenElement,
                         HTMLLabelElement, HTMLLegendElement, HTMLLIElement,
                         HTMLLinkElement, HTMLMapElement, HTMLMediaElement,
                         HTMLMetaElement, HTMLMeterElement, HTMLModElement,
                         HTMLObjectElement, HTMLOListElement,
                         HTMLOptGroupElement, HTMLOptionElement,
                         HTMLOutputElement, HTMLParagraphElement,
                         HTMLParamElement, HTMLPictureElement, HTMLPreElement,
                         HTMLProgressElement, HTMLQuoteElement,
                         HTMLScriptElement, HTMLSelectElement,
                         HTMLShadowElement, HTMLSourceElement, HTMLSpanElement,
                         HTMLStyleElement, HTMLTableCaptionElement,
                         HTMLTableCellElement, HTMLTableColElement,
                         HTMLTableDataCellElement, HTMLTableElement,
                         HTMLTableHeaderCellElement, HTMLTableRowElement,
                         HTMLTableSectionElement, HTMLTemplateElement,
                         HTMLTextAreaElement, HTMLTimeElement,
                         HTMLTitleElement, HTMLTrackElement, HTMLUListElement,
                         HTMLUnknownElement, HTMLVideoElement, Node, Text)
from domonic.webapi.url import URL

html_tags = [
    "figcaption",
    "blockquote",
    "textarea",
    "progress",
    "optgroup",
    "noscript",
    "fieldset",
    "datalist",
    "colgroup",
    "summary",
    "section",
    "details",
    "command",
    "caption",
    "article",
    "address",
    "submit",
    "strong",
    "source",
    "select",
    "script",
    "output",
    "option",
    "legend",
    "keygen",
    "iframe",
    "hgroup",
    "header",
    "footer",
    "figure",
    "canvas",
    "button",
    "video",
    "track",
    "title",
    "thead",
    "tfoot",
    "tbody",
    "table",
    "style",
    "small",
    "param",
    "meter",
    "label",
    "input",
    "audio",
    "aside",
    "applet",
    "object",
    "basefont",
    "center",
    "embed",
    "isindex",
    "listing",
    "menuitem",
    "plaintext",
    "strike",
    "template",
    "picture",
    "dialog",
    "time",
    "span",
    "samp",
    "ruby",
    "meta",
    "menu",
    "mark",
    "link",
    "html",
    "head",
    "form",
    "font",
    "code",
    "cite",
    "body",
    "base",
    "area",
    "abbr",
    "main",
    "dir",
    "wbr",
    "var",
    "sup",
    "sub",
    "nav",
    "map",
    "kbd",
    "ins",
    "img",
    "div",
    "dfn",
    "del",
    "col",
    "bdo",
    "bdi",
    "pre",
    "xmp",
    "ul",
    "tr",
    "th",
    "td",
    "rt",
    "rp",
    "ol",
    "li",
    "hr",
    "h6",
    "h5",
    "h4",
    "h3",
    "h2",
    "h1",
    "em",
    "dt",
    "dl",
    "dd",
    "br",
    "u",
    "s",
    "q",
    "p",
    "i",
    "b",
    "a",
]
# big, blink, bold, tt, var, frameset

html_attributes = [
    "accept",
    "accesskey",
    "loading",
    "action",
    "align",
    "alt",
    "async",
    "autocomplete",
    "autofocus",
    "autoplay",
    "bgcolor",
    "border",
    "charset",
    "checked",
    "cite",
    "class",
    "color",
    "cols",
    "colspan",
    "content",
    "contenteditable",
    "controls",
    "coords",
    "data",
    "datetime",
    "default",
    "defer",
    "dir",
    "dirname",
    "disabled",
    "download",
    "draggable",
    "enctype",
    "for",
    "form",
    "formaction",
    "headers",
    "height",
    "hidden",
    "high",
    "href",
    "hreflang",
    "id",
    "ismap",
    "kind",
    "label",
    "lang",
    "list",
    "loop",
    "low",
    "max",
    "maxlength",
    "media",
    "method",
    "min",
    "multiple",
    "muted",
    "name",
    "novalidate",
    "onabort",
    "onafterprint",
    "onbeforeprint",
    "onbeforeunload",
    "onblur",
    "oncanplay",
    "oncanplaythrough",
    "onchange",
    "onclick",
    "oncontextmenu",
    "oncopy",
    "oncuechange",
    "oncut",
    "ondblclick",
    "ondrag",
    "ondragend",
    "ondragenter",
    "ondragleave",
    "ondragover",
    "ondragstart",
    "ondrop",
    "ondurationchange",
    "onemptied",
    "onended",
    "onerror",
    "onfocus",
    "onhashchange",
    "oninput",
    "oninvalid",
    "onkeydown",
    "onkeypress",
    "onkeyup",
    "onload",
    "onloadeddata",
    "onloadedmetadata",
    "onloadstart",
    "onmousedown",
    "onmousemove",
    "onmouseout",
    "onmouseover",
    "onmouseup",
    "onmousewheel",
    "onoffline",
    "ononline",
    "onpagehide",
    "onpageshow",
    "onpaste",
    "onpause",
    "onplay",
    "onplaying",
    "onpopstate",
    "onprogress",
    "onratechange",
    "onreset",
    "onresize",
    "onscroll",
    "onsearch",
    "onseeked",
    "onseeking",
    "onselect",
    "onstalled",
    "onstorage",
    "onsubmit",
    "onsuspend",
    "ontimeupdate",
    "ontoggle",
    "onunload",
    "onvolumechange",
    "onwaiting",
    "onwheel",
    "open",
    "optimum",
    "pattern",
    "placeholder",
    "poster",
    "preload",
    "readonly",
    "rel",
    "required",
    "reversed",
    "rows",
    "rowspan",
    "sandbox",
    "scope",
    "selected",
    "shape",
    "size",
    "sizes",
    "span",
    "spellcheck",
    "src",
    "srcdoc",
    "srclang",
    "srcset",
    "start",
    "step",
    "style",
    "tabindex",
    "target",
    "title",
    "translate",
    "type",
    "usemap",
    "value",
    "width",
    "wrap",
    "property",
    "integrity",
    "crossorigin",
    "nonce",
    "autocapitalize",
    "enterkeyhint",
    "inputmode",
    "is",
    "itemid",
    "itemprop",
    "itemref",
    "itemscope",
    "itemtype",
    "part",
    "slot",
    "spellcheck",
    "alink",
    "nowrap",
    "vlink",
    "vspace",
    "language",
    "clear",
    "hspace",
    "xmlns",
    "about",
    "allowtransparency",
    "datatype",
    "inlist",
    "prefix",
    "resource",
    "rev",
    "typeof",
    "vocab",  # rdfa
    "playsinline",
    "autopictureinpicture",
    "buffered",
    "controlslist",
    "disableremoteplayback",  # video
]


def render(inp, outp="", to=None):  # doctype='html5')
    """write the input to string or to a file.

    Args:
        inp (obj): A domonic tag. For example div()
        outp (str): An optional output filename
        to (str): An optional output type. if 'pyml' is specified then pyml is returned instead of html.

    Returns:
        str: A HTML rendered string
    """
    if to == "pyml":
        if outp != "":
            with open(outp, "w+") as f:
                f.write(inp.__pyml__())
        return inp.__pyml__()
    # else:
    if outp != "":
        with open(outp, "w+") as f:
            f.write(str(inp))
    return str(inp)


class TemplateError(IndexError):
    def __init__(self, error, message="TemplateError: "):
        """[raised when a template error occurs]

        Args:
            error ([type]): [the error]
            message (str, optional): [description]. Defaults to "Templating error: ".
        """
        self.error = error
        self.hint = ""
        print(self.error)
        if str(self.error) == "list index out of range":
            self.hint = "MISSING UNDERSCORE ON AN ATTRIBUTE"
        self.message = message + self.hint
        super().__init__(self.message)


tag = Node  # legacy support?. TODO - remove in future? 0.0.9


class closed_tag(Node):
    def __str__(self):
        return f"<{self.name}{self.__attributes__}/>"


html = HTMLDocument
body = HTMLBodyElement
head = HTMLHeadElement
script = HTMLScriptElement
style = HTMLStyleElement
h1 = type("h1", (HTMLHeadingElement,), {"name": "h1"})
h2 = type("h2", (HTMLHeadingElement,), {"name": "h2"})
h3 = type("h3", (HTMLHeadingElement,), {"name": "h3"})
h4 = type("h4", (HTMLHeadingElement,), {"name": "h4"})
h5 = type("h5", (HTMLHeadingElement,), {"name": "h5"})
h6 = type("h6", (HTMLHeadingElement,), {"name": "h6"})
p = HTMLParagraphElement

i = type("i", (Element,), {"name": "i"})  # TODO - check which?
b = type("b", (Element,), {"name": "b"})  # TODO - check which?
portal = type("portal", (Element,), {"name": "portal"})  # TODO - check which?


def Atag(self, *args, **kwargs):

    # print('Atag: ', args, kwargs)
    # Node.__init__(self, *args, **kwargs)
    # Element.__init__(self, *args, **kwargs)

    # TODO - fix BUG. this stops having no href on a tags
    if kwargs.get("_href", None) is not None:
        URL.__init__(self, url=kwargs["_href"])
    # else:
    # Node.__init__(*args, **kwargs)
    Element.__init__(self, *args, **kwargs)
    # Node.__init__(self, *args, **kwargs)
    # URL.__init__(self, *args, **kwargs)


def __update__(
    self, *args, **kwargs
):  # TODO - you removed this but where the unit test that you wrote it for in the first place?
    # print('__update__: ', args, kwargs)
    # URL.__update__(self)
    # TODO - fix BUG. this stops having no href on a tags
    if self.getattr("_href", None) is not None:
        self.kwargs["_href"] = self.href
    # Node.__init__(self, *args, **kwargs)
    # URL.__init__(self, *args, **kwargs)
    # self.__init__(*args, **kwargs)
    Element.__init__(self, *args, **kwargs)
    URL.__init__(self, *args, **kwargs)


a = type(
    "a", (Element, URL), {"name": "a", "__init__": Atag}
)  # , "__update__": __update__})
ul = HTMLUListElement
ol = HTMLOListElement
li = HTMLLIElement
div = HTMLDivElement
strong = type("strong", (Element,), {"name": "strong"})  # TODO - check
blockquote = type("blockquote", (Element,), {"name": "blockquote"})  # TODO - check
table = HTMLTableElement
tr = type("tr", (Element,), {"name": "tr"})
td = type("td", (Element,), {"name": "td"})
# form = type('form', (Element,), {'name': 'form'})


class form(Element):
    def __init__(self, *args, **kwargs):
        new_kwargs = {}
        for k, v in kwargs.items():
            if k[0] != "_":
                # print("WARNING: kwarg '{}' should begin with an underscore".format(k))
                new_kwargs[f"_{k}"] = v
            else:
                new_kwargs[k] = v
        kwargs = new_kwargs

        self.name = "form"
        Node.__init__(self, *args, **kwargs)
        Element.__init__(self, *args, **kwargs)

    @property
    def elements(self):
        kids = []
        for child in self.children:
            if isinstance(
                child, (button, fieldset, input, object, output, select, textarea)
            ):
                kids.append(child)
        return kids


label = type("label", (Element,), {"name": "label"})
# label.__doc__ = '''
#                 .. highlight:: python
#                 .. code-block:: python

#                     # used to label form elements. i.e.
#                     label(_for=None, _text=None, **kwargs)
#                     # <label for=""></label>
#                 '''

submit = type("submit", (Element,), {"name": "submit"})
title = HTMLTitleElement
noscript = type("noscript", (Element,), {"name": "noscript"})
section = type("section", (Element,), {"name": "section"})
nav = type("nav", (Element,), {"name": "nav"})
article = type("article", (Element,), {"name": "article"})
aside = type("aside", (Element,), {"name": "aside"})
hgroup = type("hgroup", (Element,), {"name": "hgroup"})
address = type("address", (Element,), {"name": "address"})
pre = HTMLPreElement  #type("pre", (Element,), {"name": "pre"})
dl = type("dl", (Element,), {"name": "dl"})
dt = type("dt", (Element,), {"name": "dt"})
dd = type("dd", (Element,), {"name": "dd"})
figure = type("figure", (Element,), {"name": "figure"})
figcaption = type("figcaption", (Element,), {"name": "figcaption"})
em = type("em", (Element,), {"name": "em"})
small = type("small", (Element,), {"name": "small"})
s = type("s", (Element,), {"name": "s"})
cite = type("cite", (Element,), {"name": "cite"})
q = type("q", (Element,), {"name": "q"})
dfn = type("dfn", (Element,), {"name": "dfn"})
abbr = type("abbr", (Element,), {"name": "abbr"})
code = type("code", (Element,), {"name": "code"})
var = type("var", (Element,), {"name": "var"})
samp = type("samp", (Element,), {"name": "samp"})
kbd = type("kbd", (Element,), {"name": "kbd"})
sub = type("sub", (Element,), {"name": "sub"})
sup = type("sup", (Element,), {"name": "sup"})
u = type("u", (Element,), {"name": "u"})
mark = type("mark", (Element,), {"name": "mark"})
ruby = type("ruby", (Element,), {"name": "ruby"})
rt = type("rt", (Element,), {"name": "rt"})
rp = type("rp", (Element,), {"name": "rp"})
bdi = type("bdi", (Element,), {"name": "bdi"})
bdo = type("bdo", (Element,), {"name": "bdo"})
span = HTMLSpanElement
ins = type("ins", (Element,), {"name": "ins"})
iframe = type("iframe", (Element,), {"name": "iframe"})
video = HTMLVideoElement
audio = HTMLAudioElement
canvas = HTMLCanvasElement
caption = type("caption", (Element,), {"name": "caption"})
colgroup = type("colgroup", (Element,), {"name": "colgroup"})
tbody = type("tbody", (Element,), {"name": "tbody"})
thead = type("thead", (Element,), {"name": "thead"})
tfoot = type("tfoot", (Element,), {"name": "tfoot"})
th = type("th", (Element,), {"name": "th"})
fieldset = HTMLFieldSetElement  #type("fieldset", (Element,), {"name": "fieldset"})
legend = type("legend", (Element,), {"name": "legend"})
button = HTMLButtonElement  #type("button", (Element,), {"name": "button"})
select = type("select", (Element,), {"name": "select"})
datalist = HTMLDataListElement
optgroup = HTMLOptGroupElement
option = HTMLOptionElement
textarea = HTMLTextAreaElement
output = HTMLOutputElement
progress = HTMLProgressElement
meter = HTMLMeterElement
details = type("details", (Element,), {"name": "details"})
summary = type("summary", (Element,), {"name": "summary"})
menu = type("menu", (Element,), {"name": "menu"})
menuitem = type("menuitem", (Element,), {"name": "menuitem"})  # dead but may be used
font = type("font", (Element,), {"name": "font"})
header = type("header", (Element,), {"name": "header"})
footer = type("footer", (Element,), {"name": "footer"})
# map_ = type('map_', (tag,), {'name': 'map_'})
# object_ = type('object_', (tag,), {'name': 'object_'})
# del_ = type('del_', (tag,), {'name': 'del_'})
# time_ = type('time_', (tag,), {'name': 'time_'})

base = HTMLBaseElement
link = type("link", (closed_tag, Element), {"name": "link"})  # HTMLLinkElement TODO - closed tags
meta = type("meta", (closed_tag, Element), {"name": "meta"})  # HTMLMetaElement TODO - closed tags
hr = type("hr", (closed_tag, Element), {"name": "hr"})
br = type("br", (closed_tag, Element), {"name": "br"})
wbr = type("wbr", (closed_tag, Element), {"name": "wbr"})
img = type("img", (closed_tag, Element), {"name": "img"})  # HTMLImageElement TODO - closed tags
param = type("param", (closed_tag, Element), {"name": "param"})
source = type("source", (closed_tag, Element), {"name": "source"})
track = type("track", (closed_tag, Element), {"name": "track"})
area = HTMLAreaElement
col = type("col", (closed_tag, Element), {"name": "col"})
input = type("input", (closed_tag, Element), {"name": "input"})
keygen = type("keygen", (closed_tag, Element), {"name": "keygen"})
command = type("command", (closed_tag, Element), {"name": "command"})

main = type("main", (Element,), {"name": "main"})

# obsolete
applet = type("applet", (Element,), {"name": "applet"})
# object = type('object', (Element,), {'name': 'object'})
basefont = type("basefont", (Element,), {"name": "basefont"})
center = type("center", (Element,), {"name": "center"})
# dir = type('dir', (Element,), {'name': 'dir'})
embed = HTMLEmbedElement
isindex = type("isindex", (Element,), {"name": "isindex"})
listing = type("listing", (Element,), {"name": "listing"})
plaintext = type("plaintext", (Element,), {"name": "plaintext"})
s = type("s", (Element,), {"name": "s"})
u = type("u", (Element,), {"name": "u"})
strike = type("strike", (Element,), {"name": "strike"})
xmp = type("xmp", (Element,), {"name": "xmp"})

template = type("template", (Element,), {"name": "template"})

picture = HTMLPictureElement
dialog = HTMLDialogElement


# legacy.
doctype = DocumentType
comment = Comment


def create_element(name="custom_tag", *args, **kwargs):
    """
    A method for creating custom tags

    tag name needs to be set due to custom tags with hyphens can't be classnames.
    i.e. hypenated tags <some-custom-tag></some-custom-tag>
    """
    # checks if already exists
    if name in html_tags:
        return globals()[name](*args, **kwargs)

    custom_tag = type("custom_tag", (Element,), {"name": name})
    new_tag = custom_tag(*args, **kwargs)
    new_tag.name = name
    return new_tag



"""
    domonic.x3d
    ====================================
    Generate x3d with python 3

"""

from domonic.dom import Element
from domonic.html import closed_tag

x3d = X3D = type('x3d', (Element,), {'name': 'x3d'})
scene = Scene = type('scene', (Element,), {'name': 'scene'})
material = Material = type('material', (Element,), {'name': 'material'})
appearance = Appearance = type('appearance', (Element,), {'name': 'appearance'})
sphere = Sphere = type('sphere', (Element,), {'name': 'sphere'})
shape = Shape = type('shape', (Element,), {'name': 'shape'})
transform = Transform = type('transform', (Element,), {'name': 'transform'})
timeSensor = TimeSensor = type('timeSensor', (Element,), {'name': 'timeSensor'})
inline = Inline = type('inline', (Element,), {'name': 'inline'})
box = Box = type('box', (Element,), {'name': 'box'})
plane = Plane = type('plane', (Element,), {'name': 'plane'})

# TODO - go through examples to find which are usually lower. i.e before they allowed mised cases
positionInterpolator = PositionInterpolator = type('PositionInterpolator', (Element,), {'name': 'PositionInterpolator'})
route = Route = type('Route', (Element,), {'name': 'Route'})
anchor = Anchor = type('Anchor', (Element,), {'name': 'Anchor'})
arc2D = Arc2D = type('Arc2D', (Element,), {'name': 'Arc2D'})
arcClose2D = ArcClose2D = type('ArcClose2D', (Element,), {'name': 'ArcClose2D'})
audioClip = AudioClip = type('AudioClip', (Element,), {'name': 'AudioClip'})
background = Background = type('Background', (Element,), {'name': 'Background'})
ballJoint = BallJoint = type('BallJoint', (Element,), {'name': 'BallJoint'})
billboard = Billboard = type('Billboard', (Element,), {'name': 'Billboard'})
binaryGeometry = BinaryGeometry = type('BinaryGeometry', (Element,), {'name': 'BinaryGeometry'})
blendedVolumeStyle = BlendedVolumeStyle = type('BlendedVolumeStyle', (Element,), {'name': 'BlendedVolumeStyle'})
blendMode = BlendMode = type('BlendMode', (Element,), {'name': 'BlendMode'})
block = Block = type('Block', (Element,), {'name': 'Block'})
boundaryEnhancementVolumeStyle = BoundaryEnhancementVolumeStyle = type('BoundaryEnhancementVolumeStyle', (Element,), {'name': 'BoundaryEnhancementVolumeStyle'})
bufferAccessor = BufferAccessor = type('BufferAccessor', (Element,), {'name': 'BufferAccessor'})
bufferGeometry = BufferGeometry = type('BufferGeometry', (Element,), {'name': 'BufferGeometry'})
bufferView = BufferView = type('BufferView', (Element,), {'name': 'BufferView'})
cADAssembly = CADAssembly = type('CADAssembly', (Element,), {'name': 'CADAssembly'})
cADFace = CADFace = type('CADFace', (Element,), {'name': 'CADFace'})
cADLayer = CADLayer = type('CADLayer', (Element,), {'name': 'CADLayer'})
cADPart = CADPart = type('CADPart', (Element,), {'name': 'CADPart'})
cartoonVolumeStyle = CartoonVolumeStyle = type('CartoonVolumeStyle', (Element,), {'name': 'CartoonVolumeStyle'})
circle2D = Circle2D = type('Circle2D', (Element,), {'name': 'Circle2D'})
clipPlane = ClipPlane = type('ClipPlane', (Element,), {'name': 'ClipPlane'})
collidableShape = CollidableShape = type('CollidableShape', (Element,), {'name': 'CollidableShape'})
collision = Collision = type('Collision', (Element,), {'name': 'Collision'})
collisionCollection = CollisionCollection = type('CollisionCollection', (Element,), {'name': 'CollisionCollection'})
collisionSensor = CollisionSensor = type('CollisionSensor', (Element,), {'name': 'CollisionSensor'})
color = Color = type('Color', (Element,), {'name': 'Color'})
colorChaser = ColorChaser = type('ColorChaser', (Element,), {'name': 'ColorChaser'})
colorDamper = ColorDamper = type('ColorDamper', (Element,), {'name': 'ColorDamper'})
colorInterpolator = ColorInterpolator = type('ColorInterpolator', (Element,), {'name': 'ColorInterpolator'})
colorMaskMode = ColorMaskMode = type('ColorMaskMode', (Element,), {'name': 'ColorMaskMode'})
colorRGBA = ColorRGBA = type('ColorRGBA', (Element,), {'name': 'ColorRGBA'})
commonSurfaceShader = CommonSurfaceShader = type('CommonSurfaceShader', (Element,), {'name': 'CommonSurfaceShader'})
composedCubeMapTexture = ComposedCubeMapTexture = type('ComposedCubeMapTexture', (Element,), {'name': 'ComposedCubeMapTexture'})
composedShader = ComposedShader = type('ComposedShader', (Element,), {'name': 'ComposedShader'})
composedTexture3D = ComposedTexture3D = type('ComposedTexture3D', (Element,), {'name': 'ComposedTexture3D'})
composedVolumeStyle = ComposedVolumeStyle = type('ComposedVolumeStyle', (Element,), {'name': 'ComposedVolumeStyle'})
cone = Cone = type('Cone', (Element,), {'name': 'Cone'})
coordinate = Coordinate = type('Coordinate', (Element,), {'name': 'Coordinate'})
coordinateDamper = CoordinateDamper = type('CoordinateDamper', (Element,), {'name': 'CoordinateDamper'})
coordinateDouble = CoordinateDouble = type('CoordinateDouble', (Element,), {'name': 'CoordinateDouble'})
coordinateInterpolator = CoordinateInterpolator = type('CoordinateInterpolator', (Element,), {'name': 'CoordinateInterpolator'})
cylinder = Cylinder = type('Cylinder', (Element,), {'name': 'Cylinder'})
cylinderSensor = CylinderSensor = type('CylinderSensor', (Element,), {'name': 'CylinderSensor'})
depthMode = DepthMode = type('DepthMode', (Element,), {'name': 'DepthMode'})
directionalLight = DirectionalLight = type('DirectionalLight', (Element,), {'name': 'DirectionalLight'})
dish = Dish = type('Dish', (Element,), {'name': 'Dish'})
disk2D = Disk2D = type('Disk2D', (Element,), {'name': 'Disk2D'})
doubleAxisHingeJoint = DoubleAxisHingeJoint = type('DoubleAxisHingeJoint', (Element,), {'name': 'DoubleAxisHingeJoint'})
dynamicLOD = DynamicLOD = type('DynamicLOD', (Element,), {'name': 'DynamicLOD'})
edgeEnhancementVolumeStyle = EdgeEnhancementVolumeStyle = type('EdgeEnhancementVolumeStyle', (Element,), {'name': 'EdgeEnhancementVolumeStyle'})
elevationGrid = ElevationGrid = type('ElevationGrid', (Element,), {'name': 'ElevationGrid'})
environment = Environment = type('Environment', (Element,), {'name': 'Environment'})
extrusion = Extrusion = type('Extrusion', (Element,), {'name': 'Extrusion'})
field = Field = type('Field', (Element,), {'name': 'Field'})
floatVertexAttribute = FloatVertexAttribute = type('FloatVertexAttribute', (Element,), {'name': 'FloatVertexAttribute'})
fog = Fog = type('Fog', (Element,), {'name': 'Fog'})
fontStyle = FontStyle = type('FontStyle', (Element,), {'name': 'FontStyle'})
generatedCubeMapTexture = GeneratedCubeMapTexture = type('GeneratedCubeMapTexture', (Element,), {'name': 'GeneratedCubeMapTexture'})
geoCoordinate = GeoCoordinate = type('GeoCoordinate', (Element,), {'name': 'GeoCoordinate'})
geoElevationGrid = GeoElevationGrid = type('GeoElevationGrid', (Element,), {'name': 'GeoElevationGrid'})
geoLocation = GeoLocation = type('GeoLocation', (Element,), {'name': 'GeoLocation'})
geoLOD = GeoLOD = type('GeoLOD', (Element,), {'name': 'GeoLOD'})
geoMetadata = GeoMetadata = type('GeoMetadata', (Element,), {'name': 'GeoMetadata'})
geoOrigin = GeoOrigin = type('GeoOrigin', (Element,), {'name': 'GeoOrigin'})
geoPositionInterpolator = GeoPositionInterpolator = type('GeoPositionInterpolator', (Element,), {'name': 'GeoPositionInterpolator'})
geoTransform = GeoTransform = type('GeoTransform', (Element,), {'name': 'GeoTransform'})
geoViewpoint = GeoViewpoint = type('GeoViewpoint', (Element,), {'name': 'GeoViewpoint'})
group = Group = type('Group', (Element,), {'name': 'Group'})
hAnimDisplacer = HAnimDisplacer = type('HAnimDisplacer', (Element,), {'name': 'HAnimDisplacer'})
hAnimHumanoid = HAnimHumanoid = type('HAnimHumanoid', (Element,), {'name': 'HAnimHumanoid'})
hAnimJoint = HAnimJoint = type('HAnimJoint', (Element,), {'name': 'HAnimJoint'})
hAnimSegment = HAnimSegment = type('HAnimSegment', (Element,), {'name': 'HAnimSegment'})
hAnimSite = HAnimSite = type('HAnimSite', (Element,), {'name': 'HAnimSite'})
imageTexture = ImageTexture = type('ImageTexture', (Element,), {'name': 'ImageTexture'})
imageTexture3D = ImageTexture3D = type('ImageTexture3D', (Element,), {'name': 'ImageTexture3D'})
imageTextureAtlas = ImageTextureAtlas = type('ImageTextureAtlas', (Element,), {'name': 'ImageTextureAtlas'})
indexedFaceSet = IndexedFaceSet = type('IndexedFaceSet', (Element,), {'name': 'IndexedFaceSet'})
indexedLineSet = IndexedLineSet = type('IndexedLineSet', (Element,), {'name': 'IndexedLineSet'})
indexedQuadSet = IndexedQuadSet = type('IndexedQuadSet', (Element,), {'name': 'IndexedQuadSet'})
indexedTriangleSet = IndexedTriangleSet = type('IndexedTriangleSet', (Element,), {'name': 'IndexedTriangleSet'})
indexedTriangleStripSet = IndexedTriangleStripSet = type('IndexedTriangleStripSet', (Element,), {'name': 'IndexedTriangleStripSet'})
isoSurfaceVolumeData = IsoSurfaceVolumeData = type('IsoSurfaceVolumeData', (Element,), {'name': 'IsoSurfaceVolumeData'})
lineProperties = LineProperties = type('LineProperties', (Element,), {'name': 'LineProperties'})
lineSet = LineSet = type('LineSet', (Element,), {'name': 'LineSet'})
lOD = LOD = type('LOD', (Element,), {'name': 'LOD'})
matrixTextureTransform = MatrixTextureTransform = type('MatrixTextureTransform', (Element,), {'name': 'MatrixTextureTransform'})
matrixTransform = MatrixTransform = type('MatrixTransform', (Element,), {'name': 'MatrixTransform'})
mesh = Mesh = type('Mesh', (Element,), {'name': 'Mesh'})
metadataBoolean = MetadataBoolean = type('MetadataBoolean', (Element,), {'name': 'MetadataBoolean'})
metadataDouble = MetadataDouble = type('MetadataDouble', (Element,), {'name': 'MetadataDouble'})
metadataFloat = MetadataFloat = type('MetadataFloat', (Element,), {'name': 'MetadataFloat'})
metadataInteger = MetadataInteger = type('MetadataInteger', (Element,), {'name': 'MetadataInteger'})
metadataSet = MetadataSet = type('MetadataSet', (Element,), {'name': 'MetadataSet'})
metadataString = MetadataString = type('MetadataString', (Element,), {'name': 'MetadataString'})
motorJoint = MotorJoint = type('MotorJoint', (Element,), {'name': 'MotorJoint'})
movieTexture = MovieTexture = type('MovieTexture', (Element,), {'name': 'MovieTexture'})
mPRPlane = MPRPlane = type('MPRPlane', (Element,), {'name': 'MPRPlane'})
mPRVolumeStyle = MPRVolumeStyle = type('MPRVolumeStyle', (Element,), {'name': 'MPRVolumeStyle'})
multiTexture = MultiTexture = type('MultiTexture', (Element,), {'name': 'MultiTexture'})
multiTextureCoordinate = MultiTextureCoordinate = type('MultiTextureCoordinate', (Element,), {'name': 'MultiTextureCoordinate'})
navigationInfo = NavigationInfo = type('NavigationInfo', (Element,), {'name': 'NavigationInfo'})
normal = Normal = type('Normal', (Element,), {'name': 'Normal'})
normalInterpolator = NormalInterpolator = type('NormalInterpolator', (Element,), {'name': 'NormalInterpolator'})
nozzle = Nozzle = type('Nozzle', (Element,), {'name': 'Nozzle'})
opacityMapVolumeStyle = OpacityMapVolumeStyle = type('OpacityMapVolumeStyle', (Element,), {'name': 'OpacityMapVolumeStyle'})
orientationChaser = OrientationChaser = type('OrientationChaser', (Element,), {'name': 'OrientationChaser'})
orientationDamper = OrientationDamper = type('OrientationDamper', (Element,), {'name': 'OrientationDamper'})
orientationInterpolator = OrientationInterpolator = type('OrientationInterpolator', (Element,), {'name': 'OrientationInterpolator'})
orthoViewpoint = OrthoViewpoint = type('OrthoViewpoint', (Element,), {'name': 'OrthoViewpoint'})
param = Param = type('Param', (Element,), {'name': 'Param'})
particleSet = ParticleSet = type('ParticleSet', (Element,), {'name': 'ParticleSet'})
physicalEnvironmentLight = PhysicalEnvironmentLight = type('PhysicalEnvironmentLight', (Element,), {'name': 'PhysicalEnvironmentLight'})
physicalMaterial = PhysicalMaterial = type('PhysicalMaterial', (Element,), {'name': 'PhysicalMaterial'})
pixelTexture = PixelTexture = type('PixelTexture', (Element,), {'name': 'PixelTexture'})
pixelTexture3D = PixelTexture3D = type('PixelTexture3D', (Element,), {'name': 'PixelTexture3D'})
planeSensor = PlaneSensor = type('PlaneSensor', (Element,), {'name': 'PlaneSensor'})
pointLight = PointLight = type('PointLight', (Element,), {'name': 'PointLight'})
pointSet = PointSet = type('PointSet', (Element,), {'name': 'PointSet'})
polyline2D = Polyline2D = type('Polyline2D', (Element,), {'name': 'Polyline2D'})
polypoint2D = Polypoint2D = type('Polypoint2D', (Element,), {'name': 'Polypoint2D'})
popGeometry = PopGeometry = type('PopGeometry', (Element,), {'name': 'PopGeometry'})
popGeometryLevel = PopGeometryLevel = type('PopGeometryLevel', (Element,), {'name': 'PopGeometryLevel'})
positionChaser = PositionChaser = type('PositionChaser', (Element,), {'name': 'PositionChaser'})
positionChaser2D = PositionChaser2D = type('PositionChaser2D', (Element,), {'name': 'PositionChaser2D'})
positionDamper = PositionDamper = type('PositionDamper', (Element,), {'name': 'PositionDamper'})
positionDamper2D = PositionDamper2D = type('PositionDamper2D', (Element,), {'name': 'PositionDamper2D'})
positionInterpolator = PositionInterpolator = type('PositionInterpolator', (Element,), {'name': 'PositionInterpolator'})
positionInterpolator2D = PositionInterpolator2D = type('PositionInterpolator2D', (Element,), {'name': 'PositionInterpolator2D'})
projectionVolumeStyle = ProjectionVolumeStyle = type('ProjectionVolumeStyle', (Element,), {'name': 'ProjectionVolumeStyle'})
pyramid = Pyramid = type('Pyramid', (Element,), {'name': 'Pyramid'})
quadSet = QuadSet = type('QuadSet', (Element,), {'name': 'QuadSet'})
radarVolumeStyle = RadarVolumeStyle = type('RadarVolumeStyle', (Element,), {'name': 'RadarVolumeStyle'})
rectangle2D = Rectangle2D = type('Rectangle2D', (Element,), {'name': 'Rectangle2D'})
rectangularTorus = RectangularTorus = type('RectangularTorus', (Element,), {'name': 'RectangularTorus'})
refinementTexture = RefinementTexture = type('RefinementTexture', (Element,), {'name': 'RefinementTexture'})
remoteSelectionGroup = RemoteSelectionGroup = type('RemoteSelectionGroup', (Element,), {'name': 'RemoteSelectionGroup'})
renderedTexture = RenderedTexture = type('RenderedTexture', (Element,), {'name': 'RenderedTexture'})
rigidBody = RigidBody = type('RigidBody', (Element,), {'name': 'RigidBody'})
rigidBodyCollection = RigidBodyCollection = type('RigidBodyCollection', (Element,), {'name': 'RigidBodyCollection'})
scalarChaser = ScalarChaser = type('ScalarChaser', (Element,), {'name': 'ScalarChaser'})
scalarDamper = ScalarDamper = type('ScalarDamper', (Element,), {'name': 'ScalarDamper'})
scalarInterpolator = ScalarInterpolator = type('ScalarInterpolator', (Element,), {'name': 'ScalarInterpolator'})
segmentedVolumeData = SegmentedVolumeData = type('SegmentedVolumeData', (Element,), {'name': 'SegmentedVolumeData'})
shadedVolumeStyle = ShadedVolumeStyle = type('ShadedVolumeStyle', (Element,), {'name': 'ShadedVolumeStyle'})
shaderPart = ShaderPart = type('ShaderPart', (Element,), {'name': 'ShaderPart'})
silhouetteEnhancementVolumeStyle = SilhouetteEnhancementVolumeStyle = type('SilhouetteEnhancementVolumeStyle', (Element,), {'name': 'SilhouetteEnhancementVolumeStyle'})
singleAxisHingeJoint = SingleAxisHingeJoint = type('SingleAxisHingeJoint', (Element,), {'name': 'SingleAxisHingeJoint'})
sliderJoint = SliderJoint = type('SliderJoint', (Element,), {'name': 'SliderJoint'})
slopedCylinder = SlopedCylinder = type('SlopedCylinder', (Element,), {'name': 'SlopedCylinder'})
snout = Snout = type('Snout', (Element,), {'name': 'Snout'})
solidOfRevolution = SolidOfRevolution = type('SolidOfRevolution', (Element,), {'name': 'SolidOfRevolution'})
sound = Sound = type('Sound', (Element,), {'name': 'Sound'})
sphereSegment = SphereSegment = type('SphereSegment', (Element,), {'name': 'SphereSegment'})
sphereSensor = SphereSensor = type('SphereSensor', (Element,), {'name': 'SphereSensor'})
splinePositionInterpolator = SplinePositionInterpolator = type('SplinePositionInterpolator', (Element,), {'name': 'SplinePositionInterpolator'})
spotLight = SpotLight = type('SpotLight', (Element,), {'name': 'SpotLight'})
staticGroup = StaticGroup = type('StaticGroup', (Element,), {'name': 'StaticGroup'})
stippleVolumeStyle = StippleVolumeStyle = type('StippleVolumeStyle', (Element,), {'name': 'StippleVolumeStyle'})
surfaceShaderTexture = SurfaceShaderTexture = type('SurfaceShaderTexture', (Element,), {'name': 'SurfaceShaderTexture'})
switch = Switch = type('Switch', (Element,), {'name': 'Switch'})
texCoordDamper2D = TexCoordDamper2D = type('TexCoordDamper2D', (Element,), {'name': 'TexCoordDamper2D'})
text = Text = type('Text', (Element,), {'name': 'Text'})
texture = Texture = type('Texture', (Element,), {'name': 'Texture'})
textureCoordinate = TextureCoordinate = type('TextureCoordinate', (Element,), {'name': 'TextureCoordinate'})
textureCoordinate3D = TextureCoordinate3D = type('TextureCoordinate3D', (Element,), {'name': 'TextureCoordinate3D'})
textureCoordinateGenerator = TextureCoordinateGenerator = type('TextureCoordinateGenerator', (Element,), {'name': 'TextureCoordinateGenerator'})
textureProperties = TextureProperties = type('TextureProperties', (Element,), {'name': 'TextureProperties'})
textureTransform = TextureTransform = type('TextureTransform', (Element,), {'name': 'TextureTransform'})
textureTransform3D = TextureTransform3D = type('TextureTransform3D', (Element,), {'name': 'TextureTransform3D'})
textureTransformMatrix3D = TextureTransformMatrix3D = type('TextureTransformMatrix3D', (Element,), {'name': 'TextureTransformMatrix3D'})
toneMappedVolumeStyle = ToneMappedVolumeStyle = type('ToneMappedVolumeStyle', (Element,), {'name': 'ToneMappedVolumeStyle'})
torus = Torus = type('Torus', (Element,), {'name': 'Torus'})
touchSensor = TouchSensor = type('TouchSensor', (Element,), {'name': 'TouchSensor'})
triangleSet = TriangleSet = type('TriangleSet', (Element,), {'name': 'TriangleSet'})
triangleSet2D = TriangleSet2D = type('TriangleSet2D', (Element,), {'name': 'TriangleSet2D'})
twoSidedMaterial = TwoSidedMaterial = type('TwoSidedMaterial', (Element,), {'name': 'TwoSidedMaterial'})
uniform = Uniform = type('Uniform', (Element,), {'name': 'Uniform'})
universalJoint = UniversalJoint = type('UniversalJoint', (Element,), {'name': 'UniversalJoint'})
viewfrustum = Viewfrustum = type('Viewfrustum', (Element,), {'name': 'Viewfrustum'})
viewpoint = Viewpoint = type('Viewpoint', (Element,), {'name': 'Viewpoint'})
volumeData = VolumeData = type('VolumeData', (Element,), {'name': 'VolumeData'})
worldInfo = WorldInfo = type('WorldInfo', (Element,), {'name': 'WorldInfo'})
x3DAppearanceChildNode = X3DAppearanceChildNode = type('X3DAppearanceChildNode', (Element,), {'name': 'X3DAppearanceChildNode'})
x3DAppearanceNode = X3DAppearanceNode = type('X3DAppearanceNode', (Element,), {'name': 'X3DAppearanceNode'})
x3DBackgroundNode = X3DBackgroundNode = type('X3DBackgroundNode', (Element,), {'name': 'X3DBackgroundNode'})
x3DBinaryContainerGeometryNode = X3DBinaryContainerGeometryNode = type('X3DBinaryContainerGeometryNode', (Element,), {'name': 'X3DBinaryContainerGeometryNode'})
x3DBindableNode = X3DBindableNode = type('X3DBindableNode', (Element,), {'name': 'X3DBindableNode'})
x3DBoundedObject = X3DBoundedObject = type('X3DBoundedObject', (Element,), {'name': 'X3DBoundedObject'})
x3DChaserNode = X3DChaserNode = type('X3DChaserNode', (Element,), {'name': 'X3DChaserNode'})
x3DChildNode = X3DChildNode = type('X3DChildNode', (Element,), {'name': 'X3DChildNode'})
x3DColorNode = X3DColorNode = type('X3DColorNode', (Element,), {'name': 'X3DColorNode'})
x3DComposableVolumeRenderStyleNode = X3DComposableVolumeRenderStyleNode = type('X3DComposableVolumeRenderStyleNode', (Element,), {'name': 'X3DComposableVolumeRenderStyleNode'})
x3DComposedGeometryNode = X3DComposedGeometryNode = type('X3DComposedGeometryNode', (Element,), {'name': 'X3DComposedGeometryNode'})
x3DCoordinateNode = X3DCoordinateNode = type('X3DCoordinateNode', (Element,), {'name': 'X3DCoordinateNode'})
x3DDamperNode = X3DDamperNode = type('X3DDamperNode', (Element,), {'name': 'X3DDamperNode'})
x3DDragSensorNode = X3DDragSensorNode = type('X3DDragSensorNode', (Element,), {'name': 'X3DDragSensorNode'})
x3DEnvironmentNode = X3DEnvironmentNode = type('X3DEnvironmentNode', (Element,), {'name': 'X3DEnvironmentNode'})
x3DEnvironmentTextureNode = X3DEnvironmentTextureNode = type('X3DEnvironmentTextureNode', (Element,), {'name': 'X3DEnvironmentTextureNode'})
x3DFogNode = X3DFogNode = type('X3DFogNode', (Element,), {'name': 'X3DFogNode'})
x3DFollowerNode = X3DFollowerNode = type('X3DFollowerNode', (Element,), {'name': 'X3DFollowerNode'})
x3DFontStyleNode = X3DFontStyleNode = type('X3DFontStyleNode', (Element,), {'name': 'X3DFontStyleNode'})
x3DGeometricPropertyNode = X3DGeometricPropertyNode = type('X3DGeometricPropertyNode', (Element,), {'name': 'X3DGeometricPropertyNode'})
x3DGeometryNode = X3DGeometryNode = type('X3DGeometryNode', (Element,), {'name': 'X3DGeometryNode'})
x3DGroupingNode = X3DGroupingNode = type('X3DGroupingNode', (Element,), {'name': 'X3DGroupingNode'})
x3DInfoNode = X3DInfoNode = type('X3DInfoNode', (Element,), {'name': 'X3DInfoNode'})
x3DInterpolatorNode = X3DInterpolatorNode = type('X3DInterpolatorNode', (Element,), {'name': 'X3DInterpolatorNode'})
x3DLightNode = X3DLightNode = type('X3DLightNode', (Element,), {'name': 'X3DLightNode'})
x3DLODNode = X3DLODNode = type('X3DLODNode', (Element,), {'name': 'X3DLODNode'})
x3DMaterialNode = X3DMaterialNode = type('X3DMaterialNode', (Element,), {'name': 'X3DMaterialNode'})
x3DMetadataObject = X3DMetadataObject = type('X3DMetadataObject', (Element,), {'name': 'X3DMetadataObject'})
x3DNavigationInfoNode = X3DNavigationInfoNode = type('X3DNavigationInfoNode', (Element,), {'name': 'X3DNavigationInfoNode'})
x3DNBodyCollidableNode = X3DNBodyCollidableNode = type('X3DNBodyCollidableNode', (Element,), {'name': 'X3DNBodyCollidableNode'})
x3DNode = X3DNode = type('X3DNode', (Element,), {'name': 'X3DNode'})
x3DPlanarGeometryNode = X3DPlanarGeometryNode = type('X3DPlanarGeometryNode', (Element,), {'name': 'X3DPlanarGeometryNode'})
x3DPointingDeviceSensorNode = X3DPointingDeviceSensorNode = type('X3DPointingDeviceSensorNode', (Element,), {'name': 'X3DPointingDeviceSensorNode'})
x3DRigidJointNode = X3DRigidJointNode = type('X3DRigidJointNode', (Element,), {'name': 'X3DRigidJointNode'})
x3DSensorNode = X3DSensorNode = type('X3DSensorNode', (Element,), {'name': 'X3DSensorNode'})
x3DShaderNode = X3DShaderNode = type('X3DShaderNode', (Element,), {'name': 'X3DShaderNode'})
x3DShapeNode = X3DShapeNode = type('X3DShapeNode', (Element,), {'name': 'X3DShapeNode'})
x3DSoundNode = X3DSoundNode = type('X3DSoundNode', (Element,), {'name': 'X3DSoundNode'})
x3DSoundSourceNode = X3DSoundSourceNode = type('X3DSoundSourceNode', (Element,), {'name': 'X3DSoundSourceNode'})
x3DSpatialGeometryNode = X3DSpatialGeometryNode = type('X3DSpatialGeometryNode', (Element,), {'name': 'X3DSpatialGeometryNode'})
x3DTexture3DNode = X3DTexture3DNode = type('X3DTexture3DNode', (Element,), {'name': 'X3DTexture3DNode'})
x3DTextureCoordinateNode = X3DTextureCoordinateNode = type('X3DTextureCoordinateNode', (Element,), {'name': 'X3DTextureCoordinateNode'})
x3DTextureNode = X3DTextureNode = type('X3DTextureNode', (Element,), {'name': 'X3DTextureNode'})
x3DTextureTransformNode = X3DTextureTransformNode = type('X3DTextureTransformNode', (Element,), {'name': 'X3DTextureTransformNode'})
x3DTimeDependentNode = X3DTimeDependentNode = type('X3DTimeDependentNode', (Element,), {'name': 'X3DTimeDependentNode'})
x3DTouchSensorNode = X3DTouchSensorNode = type('X3DTouchSensorNode', (Element,), {'name': 'X3DTouchSensorNode'})
x3DTransformNode = X3DTransformNode = type('X3DTransformNode', (Element,), {'name': 'X3DTransformNode'})
x3DVertexAttributeNode = X3DVertexAttributeNode = type('X3DVertexAttributeNode', (Element,), {'name': 'X3DVertexAttributeNode'})
x3DViewpointNode = X3DViewpointNode = type('X3DViewpointNode', (Element,), {'name': 'X3DViewpointNode'})
x3DVolumeDataNode = X3DVolumeDataNode = type('X3DVolumeDataNode', (Element,), {'name': 'X3DVolumeDataNode'})
x3DVolumeRenderStyleNode = X3DVolumeRenderStyleNode = type('X3DVolumeRenderStyleNode', (Element,), {'name': 'X3DVolumeRenderStyleNode'})



"""
    domonic.sitemap
    ====================================

    generate or load sitemaps

    warning - when using image and video tags from this package they will be namespaced i.e <image:image> and <video:video>
    so i'd advise to only import them within the def that you use them in to avoid conflict with html.image

"""

import datetime

from domonic.dom import Document, Element

# __all__ = ['sitemap', 'url', 'lastmod']

sitemap_tags = ["sitemapindex", "sitemap", "urlset", "url", "loc", "lastmod", "changefreq", "priority",
                "image:image", "image:loc"]

sitemap_attributes = ["xmlns", "xmlns:xsi", "xsi:schemaLocation",
                      "xmlns:xhtml", "xmlns:xlink", "xmlns:atom", "xmlns:geo"]
# sitemap_change_frequencies = ['always', 'hourly', 'daily', 'weekly', 'monthly', 'yearly', 'never']

XMLNS: str = 'http://www.sitemaps.org/schemas/sitemap/0.9'
XMLNS_XSI: str = 'http://www.w3.org/2001/XMLSchema-instance'
SCHEMA_SITEINDEX: str = 'http://www.sitemaps.org/schemas/sitemap/0.9/siteindex.xsd'
SCHEMA_SITEMAP: str = 'http://www.sitemaps.org/schemas/sitemap/0.9/sitemap.xsd'
#     xmlns="http://www.google.com/schemas/sitemap-image/1.1"
#     xmlns:video="http://www.google.com/schemas/sitemap-video/1.1">


def sitemap_format(self, *args, **kwargs):
    """ attempts to prettify the output of the sitemap. """
    outp = f"<{self.name}{self.__attributes__}>{self.content}</{self.name}>"
    from domonic.xml import prettify
    return prettify(outp)


sitemapindex = type('sitemapindex', (Document,), {
    'name': 'sitemapindex',
    'xmlns': XMLNS,
    'xmlns:xsi': XMLNS_XSI,
    'xsi:schemaLocation': SCHEMA_SITEINDEX,
    '__format__': sitemap_format})

sitemap = type('sitemap', (Element,), {'name': 'sitemap'})

urlset = type('urlset', (Element,), {
    'name': 'urlset',
    'xmlns:xsi': XMLNS_XSI,
    'xsi:schemaLocation': SCHEMA_SITEMAP,
    'xmlns': XMLNS})


url = type('url', (Element,), {'name': 'url'})
loc = type('loc', (Element,), {'name': 'loc'})
lastmod = type('lastmod', (Element,), {'name': 'lastmod'})
changefreq = type('changefreq', (Element,), {'name': 'changefreq'})
priority = type('priority', (Element,), {'name': 'priority'})


def sitemapindex_from_urls(urls):
    """
    Create a sitemap index from a list of urls.

    WARNING:
    there's a difference between a sitemap index and a sitemap.
    make sure you know what you want.

    # i.e

    # <?xml version="1.0" encoding="UTF-8"?>
    # <sitemapindex xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">
    # 	<sitemap>
    # 		<loc>https://xyz.com/sitemap1.xml</loc>
    # 		<lastmod>2021-07-08T13:12:16+00:00</lastmod>
    # 	</sitemap>
    # </sitemapindex>

    """
    sitemap_index = sitemapindex()
    for url in urls:
        sitemap_index.append(sitemap(loc(url), lastmod(datetime.datetime.now())))


def sitemap_from_urls(urls):
    """
    Create a sitemap from a list of urls.add()

    Note: This won't allow you to add priority or changefreq of the urls. or add images etc
    tho u could loop the nodes afterwards and do that.

    WARNING:
    there's a difference between a sitemap index and a sitemap.
    make sure you know what you want.

    """
    sitemap = urlset()
    for url in urls:
        sitemap.append(url(loc(url), lastmod(datetime.datetime.now())))
    return sitemap


def get_sitemap(path: str, *args, **kwargs):
    # use requests to downlaod a sitemap
    import requests
    r = requests.get(path)
    import domonic
    some_sitemap = domonic.domonic.parseString(r.text)
    return some_sitemap


globals()['image:image'] = type('image:image', (Element,), {
    'name': 'image',
    'ns': 'image'})

globals()['image:loc'] = type('image:loc', (Element,), {
    'name': 'loc',
    'ns': 'image'})



"""
    domonic.utils
    ====================================
    snippets etc
"""
# import typing
import random
from collections import Counter
from itertools import chain, islice
from re import sub

from domonic.decorators import deprecated


class Utils:
    """ Utils """

    @staticmethod
    def case_camel(s: str) -> str:
        """ case_camel('camel-case') > 'camelCase' """
        s = sub(r"(_|-)+", " ", s).title().replace(" ", "")
        return s[0].lower() + s[1:]

    @staticmethod
    def case_snake(s: str) -> str:
        """
        snake('camelCase') # 'camel_case'
        """
        return '_'.join(
            sub('([A-Z][a-z]+)', r' \1',
            sub('([A-Z]+)', r' \1',
            s.replace('-', ' '))).split()).lower()

    @staticmethod
    def case_kebab(s: str) -> str:
        """
        kebab('camelCase') # 'camel-case'
        """
        return '-'.join(
            sub(r"(\s|_|-)+", " ",
            sub(r"[A-Z]{2,}(?=[A-Z][a-z]+[0-9]*|\b)|[A-Z]?[a-z]+[0-9]*|[A-Z]|[0-9]+",
            lambda mo: ' ' + mo.group(0).lower(), s)).split())

    @staticmethod
    def squash(the_list: list) -> list:
        """[turns a 2d array into a flat one]

        Args:
            the_list ([list]): [a 2d array]

        Returns:
            [list]: [a flattened 1d array]
        """
        return [inner for outer in the_list for inner in outer]

    @staticmethod
    def chunk(list: list, size: int) -> list:
        """ chunk a list into batches """
        return [list[i:i + size] for i in range(0, len(list), size)]

    @staticmethod
    def dictify(arr: list) -> dict:
        """[turns a list into a dictionary where the list items are the keys]

        Args:
            arr ([list]): [list to change]

        Returns:
            [dict]: [a new dict where the list items are now the keys]
        """
        return {}.fromkeys(arr, 0)

    @staticmethod
    def is_empty(some_str: str) -> bool:
        return (not some_str.strip())

    @staticmethod
    def unique(some_arr: list) -> list:
        """[removes duplicates from a list]

        Args:
            some_arr ([list]): [list containing duplicates]

        Returns:
            [list]: [a list containing no duplicates]
        """
        return list(set(some_arr))

    @staticmethod
    def chunks(iterable, size: int, format=iter):
        """ Iterate over any iterable (list, set, file, stream, strings, whatever), of ANY size """
        it = iter(iterable)
        while True:
            yield format(chain((it.next(),), islice(it, size - 1)))
    # >>> l = ["a", "b", "c", "d", "e", "f", "g"]
    # >>> for chunk in chunks(l, 3, tuple):
    # ...         print chunk

    @staticmethod
    def clean(lst: list) -> list:
        """[removes falsy values (False, None, 0 and “”) from a list ]

        Args:
            lst ([list]): [lst to operate on]

        Returns:
            [list]: [a new list with falsy values removed]
        """
        return list(filter(None, lst))

    @staticmethod
    def get_vowels(string: str) -> list:
        """[get a list of vowels from the word]

        Args:
            string ([str]): [the word to check]

        Returns:
            [list]: [a list of vowels]
        """
        return [each for each in string if each in 'aeiou']

    @staticmethod
    def untitle(string: str) -> str:
        """[the opposite of title]

        Args:
            str ([str]): [the string to change]

        Returns:
            [str]: [a string with the first character set to lowercase]
        """
        return string[:1].lower() + string[1:]

    @staticmethod
    def merge_dictionaries(a: dict, b: dict) -> dict:
        """[merges 2 dicts]

        Args:
            a ([dict]): [dict a]
            b ([dict]): [dict b]

        Returns:
            [dict]: [a new dict]
        """
        return {**a, **b}

    @staticmethod
    def to_dictionary(keys: list, values: list) -> dict:
        """[take a list of keys and values and returns a dict]

        Args:
            keys ([list]): [a list of keys]
            values ([list]): [a list of value]

        Returns:
            [dict]: [a dictionary]
        """
        return dict(zip(keys, values))

    @staticmethod
    def most_frequent(lst: list) -> list:
        return max(set(lst), key=lst.count)

    @staticmethod
    def is_anagram(first: str, second: str) -> bool:
        return Counter(first) == Counter(second)

    @staticmethod
    def is_palindrome(word: str) -> bool:
        return word == word[::-1]

    @staticmethod
    def acronym(sentence: str) -> str:
        """[pass a sentence, returns the acronym]

        Args:
            sentence ([str]): [typically 3 words]

        Returns:
            [str]: [a TLA (three letter acronym)]
        """
        text = sentence.split()
        a = ""
        for i in text:
            a = a + str(i[0]).upper()
        return a

    @staticmethod
    def frequency(data):
        """[check the frequency of elements in the data]

        Args:
            data ([type]): [the data to check]

        Returns:
            [dict]: [a dict of elements and their frequency]
        """
        freq = {}
        for elem in data:
            if elem in freq:
                freq[elem] += 1
            else:
                freq[elem] = 1
        return freq

    @staticmethod
    def init_assets(dir: str = 'assets') -> None:
        """[creates an assets directory with nested js/css/img dirs]

        Args:
            dir (str, optional): [default directory name]. Defaults to 'assets'.
        """
        from domonic.terminal import mkdir, touch
        mkdir(f"{dir}")
        mkdir(f"{dir}/js")
        mkdir(f"{dir}/css")
        mkdir(f"{dir}/img")
        touch(f"{dir}/js/master.js")
        touch(f"{dir}/css/style.css")
        return

    @staticmethod
    def url2file(url: str) -> str:
        """[gen a safe filename from a url. by replacing '/' for '_' and ':' for '__' ]

        Args:
            url ([str]): [the url to turn into a filename]

        Returns:
            [str]: [description]
        """
        import urllib
        url = "_".join(url.split("/"))
        url = "__".join(url.split(":"))
        filename = urllib.parse.quote_plus(url, '')
        return filename

    @staticmethod
    def permutations(word: str) -> list:
        """[provides all the possible permutations of a given word]

        Args:
            word ([str]): [the word to get permutations for]

        Returns:
            [list]: [a list of permutations]
        """
        from itertools import permutations
        return [''.join(perm) for perm in list(permutations(word))]

    @staticmethod  # TODO - remove as we have color class. (might be getting used in examples)
    def random_color(self):
        r = lambda: random.randint(0, 255)
        return str('#%02X%02X%02X' % (r(), r(), r()))

    @staticmethod
    def escape(s: str) -> str:
        """[escape a string]

        Args:
            s ([str]): [the string to escape]

        Returns:
            [str]: [the escaped string]
        """
        chars = {
            "&": "&amp;",
            '"': "&quot;",
            "'": "&apos;",
            ">": "&gt;",
            "<": "&lt;"
        }
        return "".join(chars.get(c, c) for c in s)

    @staticmethod
    def unescape(s: str) -> str:
        """[unescape a string]

        Args:
            s ([str]): [the string to unescape]

        Returns:
            [str]: [the unescaped string]
        """
        s = s.replace("&lt;", "<")
        s = s.replace("&gt;", ">")
        s = s.replace("&quot;", '"')
        s = s.replace("&apos;", "'")
        s = s.replace("&amp;", "&")
        return s

    @staticmethod
    def replace_between(content: str, match: str, replacement: str, start: int = 0, end: int = 0):
        """[replace some text but only between certain indexes]

        Args:
            content (str): [the content whos text you will be replacing]
            match (str): [the string to find]
            replacement (str): [the string to replace it with]
            start (int, optional): [start index]. Defaults to 0.
            end (int, optional): [end index]. Defaults to 0.

        Returns:
            [str]: [the new string]
        """
        front = content[0:start]
        mid = content[start:end]
        end = content[end:len(content)]
        mid = mid.replace(match, replacement)
        return front + mid + end

    @staticmethod
    def truncate(text: str = '', length: int = 0) -> str:
        """[truncates a string and appends 3 dots]

        Args:
            text (str, optional): [the text to truncate]. Defaults to ''.
            length (int, optional): [the max length]. Defaults to 0.

        Returns:
            [str]: [the truncated string]
        """
        if len(text) > length:
            return text[0:length] + "..."
        else:
            return text + "..."

    @staticmethod
    def digits(text: str = '') -> str:
        """[takes a string of mix of digits and letters and returns a string of digits]

        Args:
            text (str, optional): [the text to change]. Defaults to ''.

        Returns:
            [str]: [a string of digits]
        """
        if isinstance(text, int):
            return str(text)
        elif isinstance(text, float):
            return str(int(text))
        elif isinstance(text, str):
            return ''.join(i for i in text if i.isdigit())
        else:
            try:
                return str(text)
            except Exception:
                raise ValueError("text must be a string")

    @staticmethod
    def has_internet(url: str = 'http://www.google.com/', timeout: int = 5) -> bool:
        """[check if you have internet connection]

        Args:
            url (str, optional): [the url to check]. Defaults to 'http://www.google.com/'.
            timeout (int, optional): [the timeout]. Defaults to 5.

        Returns:
            [bool]: [True if you have internet]
        """
        import requests
        try:
            _ = requests.head(url, timeout=timeout)
            return True
        except requests.ConnectionError:
            # print("No internet connection available.")
            return False

    @staticmethod
    def is_nix() -> bool:
        """[check if the system is a nix based system]

        Returns:
            [bool]: [True if it is a nix based system]
        """
        import os
        return os.name == "posix"

    @staticmethod
    def is_mac() -> bool:
        """[check if the system is a mac]

        Returns:
            [bool]: [True if the system is a mac]
        """
        import sys
        return sys.platform == "darwin"

    @staticmethod
    def is_windows() -> bool:
        """[check if the system is a windows]

        Returns:
            [bool]: [True if windows]
        """
        import os
        return os.name == "nt"

    @staticmethod
    def is_linux() -> bool:
        """[check if the system is a linux]

        Returns:
            [bool]: [description]
        """
        import sys
        return sys.platform.startswith('linux')

    # def convert_file(filepath, filetype=None):
    #     """
    #         convert a file to a different file type
    #         mostly deals with config files
    #     """
    #  files = ['json', 'ini', 'xml', 'yaml', 'yml', 'toml', 'properties', 'conf', 'rc', 'sh', 'bash', 'bat', 'cmd', 'c', 'cpp', 'h', 'hpp', 'java', 'js', 'json', 'md', 'markdown', 'pl', 'py', 'rb', 'sh', 'sql', 'txt', 'xml', 'yaml', 'yml', 'toml']

    '''
    @staticmethod
    def yeahnah(x):
        """ returns a boolean for any given user reply
        """
        reply = x.lower()
        if reply.lower() in ['yeah', "y", "yes", "yup", "si", "yep", "yeah", "yep"]:
            return True
        elif reply.lower() in ['nah', "no", "nope", "n"]:
            return False
        else:
            # return a probability between 0 and 1 for either yes or no based on which list has the most similar words to the input
            # return max([float(reply.count(x)) / len(reply) for x in ['yeah', "y", "yes", "yup", "si", "yep", "yeah", "yep"]]) > max([float(reply.count(x)) / len(reply) for x in ['nah', "n", "no", "nope", "nop", "nope", "n", "nope"]])
            # return max([float(reply.count(x)) / len(reply) for x in ['yeah', "y", "yes", "yup", "si", "yep", "yeah", "yep"]]) > 0.5
        else:
            return None
    '''

    # def get_ip(self):
    #     """[get the current ip]

    #     Returns:
    #         [str]: [the current ip]
    #     """
    #     import socket
    #     return socket.gethostbyname(socket.gethostname())

    # def get_hostname(self):
    #     """[get the current hostname]

    #     Returns:
    #         [str]: [the current hostname]
    #     """
    #     import socket
    #     return socket.gethostname()

    # def get_mac(self):
    #     """[get the current mac]

    #     Returns:
    #         [str]: [the current mac]
    #     """
    #     import uuid
    #     return uuid.UUID(int=uuid.getnode()).hex[-12:]

    # def get_ip_mac(self):
    #     """[get the current ip and mac]

    #     Returns:
    #         [str]: [the current ip and mac]
    #     """
    #     return self.get_ip() + "|" + self.get_mac()

    # def get_os(self):
    #     """[get the current os]

    #     Returns:
    #         [str]: [the current os]
    #     """
    #     import platform
    #     return platform.system()

    # def get_os_version(self):
    #     """[get the current os version]

    #     Returns:
    #         [str]: [the current os version]
    #     """
    #     import platform
    #     return platform.release()

    # def get_os_arch(self):
    #     """[get the current os architecture]

    #     Returns:
    #         [str]: [the current os architecture]
    #     """
    #     import platform
    #     return platform.machine()

    # def get_cpu(self):
    #     """[get the current cpu]

    #     Returns:
    #         [str]: [the current cpu]
    #     """
    #     import platform
    #     return platform.processor()

    @staticmethod
    def numberToBase(n, b):
        if n == 0:
            return [0]
        digits = []
        while n:
            digits.append(int(n % b))
            n //= b
        return digits[::-1]



"""
    test_domonic
    ~~~~~~~~~~~~
    unit tests for css
"""

import json
import unittest

import domonic.JSON as JSON  # do this to use same way as previous versions of domonic
from domonic.decorators import as_json
from domonic.html import *

# import requests
# from mock import patch


class TestCase(unittest.TestCase):

    SOMEJSON = '''
    { "items": [
        {
            "id":"01",
            "name": "Java",
            "thing":{},
            "list":[]
        },
        {
            "id":"07",
            "name": "C++",
            "thing":{},
            "list":[]
        }
    ]}
    '''
    SOMEJSON2 = '''
    [
        {
            "id":"01",
            "name": "Java",
            "thing":{},
            "list":[]
        },
        {
            "id":"07",
            "name": "C++",
            "thing":{},
            "list":[]
        },
        {
            "id":"08",
            "name": "DDD",
            "thing":{},
            "list":[],
            "extra":23
        }
    ]
    '''

    def test_domonic_JSON(self):
        t = JSON.tablify(TestCase.SOMEJSON2)
        assert isinstance(t, Element)
        assert t.tagName == 'table'
        t = JSON.tablify({'id': 1, 'name': 'test'})
        assert isinstance(t, Element)
        assert t.tagName == 'table'
        t = JSON.tablify(JSON.parse(TestCase.SOMEJSON2))
        assert isinstance(t, Element)
        assert t.tagName == 'table'
        assert t.getElementsByTagName('td')[0].textContent == '01'
        t = JSON.tablify(JSON.parse(TestCase.SOMEJSON)['items'])
        assert t.getElementsByTagName('td')[0].textContent == '01'

        # t = JSON.csvify(TestCase.SOMEJSON)
        # print(t)

        # t = JSON.csv2json("data.csv")
        # print(t)

        @as_json
        def yo():
            myObj = {"hi": [1, 2, 3]}
            return myObj
        assert yo() == '{"hi": [1, 2, 3]}'

        # json_data = JSON.parse_file('surveys.json')
        # mytable = JSON.tablify(json_data)
        # print(mytable)

        # with JSON( data, 'items') as item:
        # print(item)
        # print(item.id)

        # iterator = JSON( data, 'items.age', lambda i: i<30 )

        # @JSON
        # def somefunc():
        #     returns pyobj


if __name__ == '__main__':
    unittest.main()



"""domonic - Generate html with python 3. DOM API, Javascript API and more..."""

from setuptools import find_packages, setup

from domonic import __version__ as version


def read(filename):
    """Returns the contents of a file."""
    with open(filename, encoding='utf-8') as file:
        return file.read()


def get_requirements(filename="requirements.txt"):
    """returns a list of all requirements"""
    requirements = read(filename)
    return list(
        filter(
            None,
            [req.strip() for req in requirements.split() if not req.startswith("#")],
        )
    )


setup(
    name="domonic",
    version=version,
    author="@byteface",
    author_email="byteface@gmail.com",
    license="MIT",
    url="https://github.com/byteface/domonic",
    download_url="https://github.com/byteface/domonic/archive/" + version + " .tar.gz",
    description="Generate html with python 3. DOM API, Javascript API and more...",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    keywords=[
        "html",
        "generate",
        "templating",
        "dom",
        "vdom",
        "terminal",
        "json",
        "web",
        "template",
        "javascript",
        "DOM",
        "GUI",
        "render",
        "website",
        "apps",
        "html5",
        "framework",
        "SVG",
        "x3d",
        "events",
        "geom",
    ],
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: JavaScript",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Other Audience",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Topic :: Internet",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Multimedia :: Graphics :: Presentation",
        "Topic :: Software Development",
        "Topic :: Software Development :: Code Generators",
        "Topic :: Terminals",
        "Topic :: Utilities",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing :: Markup :: HTML",
    ],
    install_requires=get_requirements(),
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "domonic = domonic.__main__:run",
        ],
    },
)
